"""Trim component.

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""

from __future__ import annotations

from typing import List, Optional, Tuple

import gdstk

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.component_layout import _parse_layer


@gf.cell
def trim(
    component: Component,
    domain: List[Tuple[float, float]],
    precision: float = 1e-4,
    return_ports: Optional[bool] = False,
) -> Component:
    """Trim a component by another geometry, preserving the component's layers and ports.

    Useful to get a smaller component from a larger one for simulation.

    Args:
        component: Component(/Reference)
        domain: list of array-like[N][2] representing the boundary of the component to keep.
        precision: float Desired precision for rounding vertex coordinates.
        return_ports: whether to return the included ports or not. Ports are always renamed to avoid inheritance conflicts.

    Returns: New component with layers (and possibly ports) of the component restricted to the domain.
    """
    domain_shape = gdstk.Polygon(domain)
    c = Component()

    for layer, layer_polygons in component.get_polygons(by_spec=True).items():
        gds_layer, gds_datatype = _parse_layer(layer)
        for layer_polygon in layer_polygons:
            p = gdstk.boolean(
                operand1=gdstk.Polygon(layer_polygon),
                operand2=domain_shape,
                operation="and",
                precision=precision,
                layer=gds_layer,
                datatype=gds_datatype,
            )
            if p:
                c.add_polygon(p, layer=layer)

    if return_ports:
        ports = []
        i = 0
        for port in component.get_ports():
            if gdstk.inside([port.center], domain_shape):
                new_name = f"{port.name[:1]}{i}"
                ports.append(port.copy(new_name))
                i += 1

        c.add_ports(ports)
        c.auto_rename_ports_layer_orientation()

    return c


if __name__ == "__main__":
    c = gf.components.straight_pin(length=10, taper=None)
    trimmed_c = trim(component=c, domain=[[0, -5], [0, 5], [5, 5], [5, -5]])
    trimmed_c.show(show_ports=True)



from __future__ import annotations

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bezier import bezier
from gdsfactory.typings import CrossSectionSpec, Float2


@cell
def bend_s(
    size: Float2 = (10.0, 2.0),
    nb_points: int = 99,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    """Return S bend with bezier curve.

    stores min_bend_radius property in self.info['min_bend_radius']
    min_bend_radius depends on height and length

    Args:
        size: in x and y direction.
        nb_points: number of points.
        cross_section: spec.
        kwargs: cross_section settings.
    """
    c = Component()
    dx, dy = size

    bend = bezier(
        control_points=((0, 0), (dx / 2, 0), (dx / 2, dy), (dx, dy)),
        npoints=nb_points,
        cross_section=cross_section,
        **kwargs,
    )
    bend_ref = c << bend
    c.add_ports(bend_ref.ports)
    c.copy_child_info(bend)
    return c


if __name__ == "__main__":
    c = bend_s(bbox_offsets=[0.5], bbox_layers=[(111, 0)], width=2)
    # c = bend_s(size=[10, 2.5])  # 10um bend radius
    # c = bend_s(size=[20, 3], cross_section="rib")  # 10um bend radius
    # c.pprint()
    # c = bend_s_biased()
    # print(c.info["min_bend_radius"])
    c.show(show_ports=False)



"""In programming, a factory is a function that returns an object.

Functions are easy to understand because they have clear inputs and outputs.
Most gdsfactory functions take some inputs and return a Component object.
Some of these inputs parameters are also functions.

- Component: Object with.
    - name.
    - references: to other components (x, y, rotation).
    - polygons in different layers.
    - ports dict.
- Route: dataclass with 3 attributes.
    - references: list of references (straights, bends and tapers).
    - ports: dict(input=PortIn, output=PortOut).
    - length: how long is this route?

Factories:

- ComponentFactory: function that returns a Component.
- RouteFactory: function that returns a Route.


Specs:

- ComponentSpec: Component, function, string or dict
    (component=mzi, settings=dict(delta_length=20)).
- LayerSpec: (3, 0), 3 (assumes 0 as datatype) or string.

"""
from __future__ import annotations

import json
import pathlib
import dataclasses
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import gdstk
import numpy as np
from gdstk import Label
from omegaconf import OmegaConf
from pydantic import BaseModel, Extra
from typing_extensions import Literal

from gdsfactory.component import Component, ComponentReference
from gdsfactory.cross_section import CrossSection, Section
from gdsfactory.port import Port
from gdsfactory.technology import LayerLevel, LayerStack


STEP_DIRECTIVES = {
    "x",
    "y",
    "dx",
    "dy",
}

STEP_DIRECTIVES_ALL_ANGLE = {
    "x",
    "y",
    "dx",
    "dy",
    "ds",
    "exit_angle",
    "cross_section",
    "connector",
    "separation",
}


@dataclasses.dataclass
class Step:
    """Manhattan Step.

    Parameters:
        x: absolute.
        y: absolute.
        dx: x-displacement.
        dy: y-displacement.

    """

    x: Optional[float] = None
    y: Optional[float] = None
    dx: Optional[float] = None
    dy: Optional[float] = None


@dataclasses.dataclass
class StepAllAngle:
    x: Optional[float] = None
    y: Optional[float] = None
    dx: Optional[float] = None
    dy: Optional[float] = None
    ds: Optional[float] = None
    exit_angle: Optional[float] = None
    cross_section: Optional[CrossSectionSpec] = None
    connector: Optional[ComponentSpec] = None
    separation: Optional[float] = None

    """All angle Ste.

    Parameters:
        x: absolute.
        y: absolute.
        dx: x-displacement.
        dy: y-displacement.
        exit_angle: in degrees.
        cross_section: spec.
        connector: define transition.
        separation: in um.

    """

    class Config:
        """Config for Steps with all angle."""

        extra = Extra.forbid


Anchor = Literal[
    "ce",
    "cw",
    "nc",
    "ne",
    "nw",
    "sc",
    "se",
    "sw",
    "center",
    "cc",
]
Axis = Literal["x", "y"]
NSEW = Literal["N", "S", "E", "W"]
WidthTypes = Literal["sine", "linear", "parabolic"]


Float2 = Tuple[float, float]
Float3 = Tuple[float, float, float]
Floats = Tuple[float, ...]
Strs = Tuple[str, ...]
Int2 = Tuple[int, int]
Int3 = Tuple[int, int, int]
Ints = Tuple[int, ...]

Layer = Tuple[int, int]  # Tuple of integer (layer, datatype)
Layers = Tuple[Layer, ...]

LayerSpec = Union[
    Layer, int, str, None
]  # tuple of integers (layer, datatype), a integer (layer, 0) or a string (layer_name)

LayerSpecs = Optional[Tuple[LayerSpec, ...]]
ComponentFactory = Callable[..., Component]
ComponentFactoryDict = Dict[str, ComponentFactory]
PathType = Union[str, pathlib.Path]
PathTypes = Tuple[PathType, ...]

ComponentOrPath = Union[PathType, Component]
ComponentOrReference = Union[Component, ComponentReference]
NameToFunctionDict = Dict[str, ComponentFactory]
Number = Union[float, int]
Coordinate = Tuple[float, float]
Coordinates = Tuple[Coordinate, ...]
ComponentOrPath = Union[Component, PathType]
CrossSectionFactory = Callable[..., CrossSection]
CrossSectionOrFactory = Union[CrossSection, Callable[..., CrossSection]]
PortSymmetries = Dict[str, List[str]]
PortsDict = Dict[str, Port]
PortsList = Dict[str, Port]

ComponentSpec = Union[
    str, ComponentFactory, Component, Dict[str, Any]
]  # PCell function, function name, dict or Component

ComponentSpecOrList = Union[ComponentSpec, List[ComponentSpec]]
CellSpec = Union[
    str, ComponentFactory, Dict[str, Any]
]  # PCell function, function name or dict

ComponentSpecDict = Dict[str, ComponentSpec]
CrossSectionSpec = Union[
    str, CrossSectionFactory, CrossSection, Dict[str, Any]
]  # cross_section function, function name or dict

MultiCrossSectionAngleSpec = List[Tuple[CrossSectionSpec, Tuple[int, ...]]]


class Route(BaseModel):
    references: List[ComponentReference]
    labels: Optional[List[gdstk.Label]] = None
    ports: Tuple[Port, Port]
    length: float

    class Config:
        """Config for Route."""

        extra = Extra.forbid
        arbitrary_types_allowed = True


class Routes(BaseModel):
    references: List[ComponentReference]
    lengths: List[float]
    ports: Optional[List[Port]] = None
    bend_radius: Optional[List[float]] = None

    class Config:
        """Config for Routes."""

        extra = Extra.forbid


class ComponentModel(BaseModel):
    component: Union[str, Dict[str, Any]]
    settings: Optional[Dict[str, Any]]

    class Config:
        extra = Extra.forbid


class PlacementModel(BaseModel):
    x: Union[str, float] = 0
    y: Union[str, float] = 0
    xmin: Optional[Union[str, float]] = None
    ymin: Optional[Union[str, float]] = None
    xmax: Optional[Union[str, float]] = None
    ymax: Optional[Union[str, float]] = None
    dx: float = 0
    dy: float = 0
    port: Optional[Union[str, Anchor]] = None
    rotation: int = 0
    mirror: bool = False

    class Config:
        extra = Extra.forbid


class RouteModel(BaseModel):
    links: Dict[str, str]
    settings: Optional[Dict[str, Any]] = None
    routing_strategy: Optional[str] = None

    class Config:
        extra = Extra.forbid


class NetlistModel(BaseModel):
    """Netlist defined component.

    Parameters:
        instances: dict of instances (name, settings, component).
        placements: dict of placements.
        connections: dict of connections.
        routes: dict of routes.
        name: component name.
        info: information (polarization, wavelength ...).
        settings: input variables.
        ports: exposed component ports.

    """

    instances: Optional[Dict[str, ComponentModel]] = None
    placements: Optional[Dict[str, PlacementModel]] = None
    connections: Optional[Dict[str, str]] = None
    routes: Optional[Dict[str, RouteModel]] = None
    name: Optional[str] = None
    info: Optional[Dict[str, Any]] = None
    settings: Optional[Dict[str, Any]] = None
    ports: Optional[Dict[str, str]] = None

    class Config:
        extra = Extra.forbid


RouteFactory = Callable[..., Route]


class TypedArray(np.ndarray):
    """based on https://github.com/samuelcolvin/pydantic/issues/380."""

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_type

    @classmethod
    def validate_type(cls, val):
        return np.array(val, dtype=cls.inner_type)


class ArrayMeta(type):
    def __getitem__(self, t):
        return type("Array", (TypedArray,), {"inner_type": t})


class Array(np.ndarray, metaclass=ArrayMeta):
    pass


__all__ = (
    "Callable",
    "Component",
    "ComponentFactory",
    "ComponentFactoryDict",
    "ComponentSpec",
    "ComponentOrPath",
    "ComponentOrReference",
    "Coordinate",
    "Coordinates",
    "CrossSection",
    "CrossSectionFactory",
    "CrossSectionOrFactory",
    "CrossSectionSpec",
    "MultiCrossSectionAngleSpec",
    "Float2",
    "Float3",
    "Floats",
    "Int2",
    "Int3",
    "Ints",
    "Layer",
    "Label",
    "Layers",
    "LayerLevel",
    "LayerStack",
    "LayerSpec",
    "LayerSpecs",
    "NameToFunctionDict",
    "Number",
    "PathType",
    "PathTypes",
    "Route",
    "RouteFactory",
    "Routes",
    "Strs",
    "Section",
    "Any",
    "Dict",
    "List",
    "Optional",
    "Union",
    "Tuple",
)


def write_schema(model: BaseModel = NetlistModel) -> None:
    from gdsfactory.config import PATH

    s = model.schema_json()
    d = OmegaConf.create(s)

    schema_path_json = PATH.schema_netlist
    schema_path_yaml = schema_path_json.with_suffix(".yaml")

    schema_path_yaml.write_text(OmegaConf.to_yaml(d))
    schema_path_json.write_text(json.dumps(OmegaConf.to_container(d)))


def _demo():
    write_schema()

    import jsonschema
    import yaml

    from gdsfactory.config import PATH

    schema_path_json = PATH.schema_netlist
    schema_dict = json.loads(schema_path_json.read_text())

    yaml_text = """

name: mzi

pdk: ubcpdk

settings:
   dy: -90

info:
    polarization: te
    wavelength: 1.55
    description: mzi for ubcpdk

instances:
    yr:
      component: y_splitter
    yl:
      component: y_splitter

placements:
    yr:
        rotation: 180
        x: 100
        y: 0

routes:
    route_top:
        links:
            yl,opt2: yr,opt3
        settings:
            cross_section: strip
    route_bot:
        links:
            yl,opt3: yr,opt2
        routing_strategy: get_bundle_from_steps
        settings:
          steps: [dx: 30, dy: '${settings.dy}', dx: 20]
          cross_section: strip

ports:
    o1: yl,opt1
    o2: yr,opt1
"""

    yaml_dict = yaml.safe_load(yaml_text)
    jsonschema.validate(yaml_dict, schema_dict)

    # from gdsfactory.components import factory
    # c = NetlistModel(factory=factory)
    # c.add_instance("mmi1", "mmi1x2", length=13.3)


if __name__ == "__main__":
    # _demo()
    s = Step()



"""fill is now slower.

This code takes now 5 seconds to run.

in version 5.18.6 it was taking 0.66 seconds to run
"""

if __name__ == "__main__":
    import time

    import gdsfactory as gf

    start = time.time()

    coupler_lengths = [10, 20, 30, 40, 50, 60, 70, 80]
    coupler_gaps = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]
    delta_lengths = [10, 100, 200, 300, 400, 500, 500]

    coupler_lengths = [10, 20]
    coupler_gaps = [0.1, 0.2]
    delta_lengths = [10]

    n = 4
    coupler_lengths = [10] * n
    coupler_gaps = [0.2] * n
    delta_lengths = [10] * (n - 1)

    mzi = gf.components.mzi_lattice(
        coupler_lengths=coupler_lengths,
        coupler_gaps=coupler_gaps,
        delta_lengths=delta_lengths,
    )

    # Add fill
    c = gf.Component("component_with_fill")
    layers = [(1, 0)]
    fill_size = [0.5, 0.5]

    c << gf.fill_rectangle(
        mzi,
        fill_size=fill_size,
        fill_layers=layers,
        margin=5,
        fill_densities=[0.8] * len(layers),
        avoid_layers=layers,
    )
    c << mzi
    c.show(show_ports=True)
    end = time.time()
    print(end - start)



from __future__ import annotations

import inspect
from typing import List, Tuple

import gdsfactory as gf
from gdsfactory.add_labels import get_input_label
from gdsfactory.cell import cell
from gdsfactory.component import Component, ComponentReference
from gdsfactory.components.grating_coupler_elliptical_trenches import grating_coupler_te
from gdsfactory.port import Port
from gdsfactory.routing.get_route import get_route
from gdsfactory.routing.manhattan import round_corners
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


def connect_loopback(
    port0: Port,
    port1: Port,
    a: float,
    b: float,
    y_bot_align_route: float,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> List[ComponentReference]:
    p0 = port0.center
    p1 = port1.center
    points = [
        p0,
        p0 + (0, a),
        p0 + (b, a),
        p0 + (b, y_bot_align_route),
        p1 + (-b, y_bot_align_route),
        p1 + (-b, a),
        p1 + (0, a),
        p1,
    ]

    bend90 = gf.components.bend_euler(cross_section=cross_section, **kwargs)
    return round_corners(
        points=points, bend=bend90, straight=gf.components.straight, **kwargs
    ).references


@cell
def loss_deembedding_ch13_24(
    pitch: float = 127.0,
    grating_coupler: ComponentSpec = grating_coupler_te,
    input_port_indexes: Tuple[int, ...] = (0, 1),
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Component:
    """Grating coupler test structure for fiber array.

    Connects channel 1->3, 2->4

    Args:
        pitch: um.
        grating_coupler: spec.
        input_port_indexes: adds test labels.
        cross_section: spec.
        kwargs: cross_section settings.
    """
    gc = gf.get_component(grating_coupler)
    c = gf.Component()
    dx = pitch
    gcs = [gc.ref(position=(i * dx, 0), port_id="o1", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["o1"] for g in gcs]
    c.add(gcs)

    c.add(
        get_route(
            gc_ports[0],
            gc_ports[2],
            start_straight_length=40.0,
            taper=None,
            cross_section=cross_section,
            **kwargs,
        ).references
    )

    x = gf.get_cross_section(cross_section, **kwargs)
    radius = x.radius

    gsi = gc.size_info
    p1 = gc_ports[1]
    p3 = gc_ports[3]
    a = radius + 5.0  # 0.5
    b = max(2 * a, pitch / 2)
    y_bot_align_route = -gsi.width - 5.0

    c.add(
        connect_loopback(
            p1, p3, a, b, y_bot_align_route, cross_section=cross_section, **kwargs
        )
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.origin = gc_ports[index].center
        c.add(label)

    return c


@cell
def loss_deembedding_ch12_34(
    pitch: float = 127.0,
    grating_coupler: ComponentSpec = grating_coupler_te,
    input_port_indexes: Tuple[int, ...] = (0, 2),
    **kwargs,
) -> Component:
    """Grating coupler test structure for fiber array.

    Connects channel 1->2, 3->4

    Args:
        pitch: um.
        grating_coupler: spec.
        input_port_indexes: for grating couplers.

    Keyword Args:
        cross_section: spec.
        kwargs: cross_section settings.
    """
    gc = gf.get_component(grating_coupler)

    c = gf.Component()
    dx = pitch
    gcs = [gc.ref(position=(i * dx, 0), port_id="o1", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["o1"] for g in gcs]
    c.add(gcs)

    c.add(
        get_route(
            gc_ports[0], gc_ports[1], start_straight_length=40.0, taper=None, **kwargs
        ).references
    )
    c.add(
        get_route(
            gc_ports[2], gc_ports[3], start_straight_length=40.0, taper=None, **kwargs
        ).references
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.origin = gc_ports[index].center
        c.add(label)
    return c


@cell
def loss_deembedding_ch14_23(
    pitch: float = 127.0,
    grating_coupler: ComponentSpec = grating_coupler_te,
    input_port_indexes: Tuple[int, ...] = (0, 1),
    **kwargs,
) -> Component:
    """Grating coupler test structure for fiber array.

    Connects channel 1->4, 2->3

    Args:
        pitch: um.
        grating_coupler: spec.
        input_port_indexes: for grating couplers.

    Keyword Args:
        cross_section: spec.
        kwargs: cross_section settings.
    """
    gc = gf.get_component(grating_coupler)

    c = gf.Component()
    dx = pitch
    gcs = [gc.ref(position=(i * dx, 0), port_id="o1", rotation=-90) for i in range(4)]

    gc_ports = [g.ports["o1"] for g in gcs]
    c.add(gcs)

    c.add(
        get_route(
            gc_ports[0], gc_ports[3], start_straight_length=40.0, taper=None, **kwargs
        ).references
    )
    c.add(
        get_route(
            gc_ports[1], gc_ports[2], start_straight_length=30.0, taper=None, **kwargs
        ).references
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.origin = gc_ports[index].center
        c.add(label)
    return c


@cell
def grating_coupler_loss_fiber_array(
    pitch: float = 127.0,
    grating_coupler: ComponentSpec = grating_coupler_te,
    input_port_indexes: Tuple[int, ...] = (0, 1),
    **kwargs,
) -> Component:
    """Returns Grating coupler fiber array loopback.

    Args:
        pitch: spacing.
        grating_coupler: spec for grating coupler.
        input_port_indexes: for grating couplers.

    Keyword Args:
        cross_section: spec.
        kwargs: cross_section settings.
    """
    gc = gf.get_component(grating_coupler)

    c = gf.Component()
    dx = pitch
    gcs = [gc.ref(position=(i * dx, 0), port_id="o1", rotation=-90) for i in range(2)]

    gc_ports = [g.ports["o1"] for g in gcs]
    c.add(gcs)

    c.add(
        get_route(
            gc_ports[0], gc_ports[1], start_straight_length=40.0, taper=None, **kwargs
        ).references
    )
    for i, index in enumerate(input_port_indexes):
        label = get_input_label(
            gc_ports[index], gc, i, component_name=inspect.stack()[0][3]
        )
        label.origin = gc_ports[index].center
        c.add(label)
    return c


@cell
def grating_coupler_loss_fiber_array4(
    pitch: float = 127.0, grating_coupler: ComponentSpec = grating_coupler_te, **kwargs
) -> Component:
    """Returns a grating coupler test structure for fiber array.

    Measures all combinations for a 4 fiber fiber_array

    Connects channel 1->3, 2->4
    Connects channel 1->4, 2->3
    Connects channel 1->2, 3->4

    Args:
        pitch: grating_coupler_pitch.
        grating_coupler: function.
        kwargs: cross_section settings.
    """
    c = gf.Component()
    c1 = loss_deembedding_ch13_24(grating_coupler=grating_coupler, **kwargs)
    c2 = loss_deembedding_ch14_23(grating_coupler=grating_coupler, **kwargs)
    c3 = loss_deembedding_ch12_34(grating_coupler=grating_coupler, **kwargs)
    c.add_ref(c1)
    c2 = c.add_ref(c2)
    c3 = c.add_ref(c3)
    c2.movex(pitch * 4)
    c3.movex(pitch * 8)

    return c


if __name__ == "__main__":
    # c = loss_deembedding_ch14_23()
    # c = loss_deembedding_ch12_34()
    # c = loss_deembedding_ch13_24()
    c = grating_coupler_loss_fiber_array4(layer=(2, 0), radius=30)
    # c = grating_coupler_loss_fiber_array(layer=(2, 0), radius=30)
    c.show(show_ports=True)



from __future__ import annotations

from typing import Dict, Optional

import gmsh

from gdsfactory.simulation.gmsh.parse_component import buffers_to_lists
from gdsfactory.simulation.gmsh.parse_gds import cleanup_component
from gdsfactory.simulation.gmsh.parse_layerstack import order_layerstack
from gdsfactory.technology import LayerStack
from gdsfactory.typings import ComponentOrReference


def add_get_point(occ, x, y, z, points_dict):
    """Add a point to the gmsh model, or retrieve a previously-defined point.

    Args:
        model: GMSH model
        x: float, x-coordinate
        y: float, y-coordinate
        z: float, z-coordinate
    """
    if (x, y, z) not in points_dict.keys():
        points_dict[(x, y, z)] = occ.add_point(x, y, z)
    return points_dict[(x, y, z)]


def add_get_segment(occ, xyz1, xyz2, lines_dict, points_dict):
    """Add a segment (2-point line) to the gmsh model, or retrieve a previously-defined segment.

    Note that the OCC kernel does not care about orientation.

    Args:
        model: GMSH model
        xyz1: first [x,y,z] coordinate
        xyz2: second [x,y,z] coordinate
    """
    if (xyz1, xyz2) in lines_dict.keys():
        return lines_dict[(xyz1, xyz2)]
    elif (xyz2, xyz1) in lines_dict.keys():
        return lines_dict[(xyz2, xyz1)]
    else:
        lines_dict[(xyz1, xyz2)] = occ.add_line(
            add_get_point(occ, xyz1[0], xyz1[1], xyz1[2], points_dict),
            add_get_point(occ, xyz2[0], xyz2[1], xyz2[2], points_dict),
        )
        return lines_dict[(xyz1, xyz2)]


def channel_loop_from_vertices(occ, vertices, lines_dict, points_dict):
    """Add a curve loop from the list of vertices.

    Args:
        model: GMSH model
        vertices: list of [x,y,z] coordinates
    """
    edges = []
    for vertex1, vertex2 in [
        (vertices[i], vertices[i + 1]) for i in range(len(vertices) - 1)
    ]:
        gmsh_line = add_get_segment(occ, vertex1, vertex2, lines_dict, points_dict)
        edges.append(gmsh_line)
    return occ.add_curve_loop(edges)


def add_surface(occ, vertices, lines_dict, points_dict):
    """Add a surface composed of the segments formed by vertices.

    Args:
        vertices: List of xyz coordinates, whose subsequent entries define a closed loop.
    """
    channel_loop = channel_loop_from_vertices(occ, vertices, lines_dict, points_dict)
    return occ.add_plane_surface([channel_loop])


def add_volume(occ, entry, lines_dict, points_dict, exterior=True, interior_index=0):
    """Create shape from a list of the same buffered polygon and a list of z-values.

    Args:
        polygons: shapely polygons from the GDS
        zs: list of z-values for each polygon

    Returns:
        GMSH volume for this entry.
    """
    # Draw bottom surface
    bottom_polygon = entry[0][1]
    bottom_polygon_z = entry[0][0]
    if exterior:
        bottom_polygon_vertices = [
            (x, y, bottom_polygon_z) for x, y in bottom_polygon.exterior.coords
        ]
    else:
        bottom_polygon_vertices = [
            (x, y, bottom_polygon_z)
            for x, y in bottom_polygon.interiors[interior_index].coords
        ]
    gmsh_surfaces = [add_surface(occ, bottom_polygon_vertices, lines_dict, points_dict)]
    # Draw top surface
    top_polygon = entry[-1][1]
    top_polygon_z = entry[-1][0]
    if exterior:
        top_polygon_vertices = [
            (x, y, top_polygon_z) for x, y in top_polygon.exterior.coords
        ]
    else:
        top_polygon_vertices = [
            (x, y, top_polygon_z)
            for x, y in top_polygon.interiors[interior_index].coords
        ]
    gmsh_surfaces.append(
        add_surface(occ, top_polygon_vertices, lines_dict, points_dict)
    )
    # Draw vertical surfaces
    for pair_index in range(len(entry) - 1):
        if exterior:
            bottom_polygon = entry[pair_index][1].exterior.coords
            top_polygon = entry[pair_index + 1][1].exterior.coords
        else:
            bottom_polygon = entry[pair_index][1].interiors[interior_index].coords
            top_polygon = entry[pair_index + 1][1].interiors[interior_index].coords
        bottom_z = entry[pair_index][0]
        top_z = entry[pair_index + 1][0]
        for facet_pt_ind in range(len(bottom_polygon) - 1):
            facet_pt1 = (
                bottom_polygon[facet_pt_ind][0],
                bottom_polygon[facet_pt_ind][1],
                bottom_z,
            )
            facet_pt2 = (
                bottom_polygon[facet_pt_ind + 1][0],
                bottom_polygon[facet_pt_ind + 1][1],
                bottom_z,
            )
            facet_pt3 = (
                top_polygon[facet_pt_ind + 1][0],
                top_polygon[facet_pt_ind + 1][1],
                top_z,
            )
            facet_pt4 = (
                top_polygon[facet_pt_ind][0],
                top_polygon[facet_pt_ind][1],
                top_z,
            )
            facet_vertices = [facet_pt1, facet_pt2, facet_pt3, facet_pt4, facet_pt1]
            gmsh_surfaces.append(
                add_surface(occ, facet_vertices, lines_dict, points_dict)
            )

    # Return volume from closed shell
    surface_loop = occ.add_surface_loop(gmsh_surfaces)
    return occ.add_volume([surface_loop])


def add_volume_with_holes(occ, entry, lines_dict, points_dict):
    """Returns volume, removing intersection with hole volumes."""
    exterior = add_volume(occ, entry, lines_dict, points_dict, exterior=True)
    interiors = [
        add_volume(
            occ,
            entry,
            lines_dict,
            points_dict,
            exterior=False,
            interior_index=interior_index,
        )
        for interior_index in range(len(entry[0][1].interiors))
    ]
    if interiors:
        for interior in interiors:
            exterior = occ.cut(
                [(3, exterior)], [(3, interior)], removeObject=True, removeTool=True
            )
            occ.synchronize()
            exterior = exterior[0][0][1]  # Parse `outDimTags', `outDimTagsMap'
    return exterior


def create_shapes(occ, buffered_layer_polygons_dict):
    """Loop over layers and polygons to create base shapes.

    Args:
        buffered_layer_polygons_dict:

    Returns:
        Dict of layernames: occ volume ids
    """
    shapes = {}
    lines_dict = {}
    points_dict = {}
    for layername, entry in buffered_layer_polygons_dict.items():
        subshapes_list = []
        subshapes_list = [
            add_volume_with_holes(occ, polygon, lines_dict, points_dict)
            for polygon in entry
        ]
        shapes[layername] = subshapes_list

    return shapes


def xyz_mesh(
    component: ComponentOrReference,
    layerstack: LayerStack,
    resolutions: Optional[Dict] = None,
    default_resolution_min: float = 0.01,
    default_resolution_max: float = 0.5,
    filename: Optional[str] = None,
    verbosity: Optional[bool] = False,
    override_volumes: Optional[Dict] = None,
    round_tol: int = 3,
    simplify_tol: float = 1e-2,
):
    """Full 3D mesh of component.

    Args:
        component (Component): gdsfactory component to mesh
        layerstack (LayerStack): gdsfactory LayerStack to parse
        resolutions (Dict): Pairs {"layername": {"resolution": float, "distance": "float}} to roughly control mesh refinement
        default_resolution_min (float): gmsh minimal edge length
        default_resolution_max (float): gmsh maximal edge length
        filename (str, path): where to save the .msh file
        override_volumes: Dict of {physical: [volume_ids]}. If not None, will manually assign physicals to the volume IDs (after performing coherence), deleting extra volumes
        round_tol: during gds --> mesh conversion cleanup, number of decimal points at which to round the gdsfactory/shapely points before introducing to gmsh
        simplify_tol: during gds --> mesh conversion cleanup, shapely "simplify" tolerance (make it so all points are at least separated by this amount)
    """
    # Fuse and cleanup polygons of same layer in case user overlapped them
    layer_polygons_dict = cleanup_component(
        component, layerstack, round_tol, simplify_tol
    )

    # GDS polygons to simulation polygons
    buffered_layer_polygons_dict = buffers_to_lists(layer_polygons_dict, layerstack)

    occ = gmsh.model.occ
    gmsh.initialize()
    gmsh.clear()
    gmsh.option.setNumber("Geometry.OCCBooleanPreserveNumbering", 1)

    shapes = create_shapes(occ, buffered_layer_polygons_dict)
    occ.synchronize()

    # Iterate through objects, removing overlaps
    # Don't remove entities at this stage
    broken_shapes = {}
    ordered_layers = order_layerstack(layerstack)  # gds layers
    for index, layername in enumerate(ordered_layers):
        if index == 0:
            broken_shapes[layername] = shapes[layername]
            tool = shapes[layername]
        else:
            new_obj = []
            for obj_id in shapes[layername]:
                new_tool = []
                for tool_id in tool:
                    new_obj.append(
                        [
                            x[1]
                            for x in occ.cut(
                                [(3, obj_id)],
                                [(3, tool_id)],
                                removeObject=False,
                                removeTool=False,
                            )[0]
                        ]
                        or [obj_id]
                    )
                    new_tool.append(
                        [
                            x[1]
                            for x in occ.fuse(
                                [(3, obj_id)],
                                [(3, tool_id)],
                                removeObject=False,
                                removeTool=False,
                            )[0]
                        ]
                    )
            tool = [item for sublist in new_tool for item in sublist]
            broken_shapes[layername] = [
                item for sublist in new_obj for item in sublist if item
            ]

    # Remove redundant entities
    occ.synchronize()
    current_entities = [item for sublist in broken_shapes.values() for item in sublist]
    entities_to_remove = [
        (dimension, entity)
        for dimension, entity in occ.getEntities(dim=3)
        if entity not in current_entities
    ]
    occ.remove(entities_to_remove, recursive=False)
    occ.synchronize()

    # Fix remaining interfaces
    # HACK: Assuming GMSH iterates in entity order, reassing entities by relative order if they changed after duplicate deletion
    # Might be more robust if we re-identify with something calculated from the entity like center and bounding box
    pre_removeAllDuplicates = {entity for dimension, entity in occ.getEntities(dim=3)}
    occ.removeAllDuplicates()
    occ.synchronize()
    post_removeAllDuplicates = {entity for dimension, entity in occ.getEntities(dim=3)}

    if override_volumes:
        for layername, volume_ids in override_volumes.items():
            gmsh.model.addPhysicalGroup(3, volume_ids, name=layername)
        current_entities = [
            item for sublist in override_volumes.values() for item in sublist
        ]
        entities_to_remove = [
            (dimension, entity)
            for dimension, entity in occ.getEntities(dim=3)
            if entity not in current_entities
        ]
        occ.remove(entities_to_remove, recursive=False)
        occ.synchronize()

    else:  # try to smartly reassign volumes
        old_entities = list(pre_removeAllDuplicates - post_removeAllDuplicates)
        new_entities = list(post_removeAllDuplicates - pre_removeAllDuplicates)

        for layername, layer_old_entities in broken_shapes.items():
            layer_new_entities = []
            for entity in layer_old_entities:
                if entity in old_entities:
                    layer_new_entities.append(new_entities[old_entities.index(entity)])
                else:
                    layer_new_entities.append(entity)
            broken_shapes[layername] = layer_new_entities

        for layername, entities in broken_shapes.items():
            gmsh.model.addPhysicalGroup(3, entities, name=layername)

    # Refine
    n = 0
    refinement_fields = []
    for label, resolution in resolutions.items():
        # Inside surface
        mesh_resolution = resolution["resolution"]
        gmsh.model.mesh.field.add("MathEval", n)
        gmsh.model.mesh.field.setString(n, "F", f"{mesh_resolution}")
        gmsh.model.mesh.field.add("Restrict", n + 1)
        gmsh.model.mesh.field.setNumber(n + 1, "InField", n)
        gmsh.model.mesh.field.setNumbers(
            n + 1,
            "RegionsList",
            broken_shapes[label],
        )
        refinement_fields.extend((n + 1,))
        n += 2

    # Use the smallest element size overall
    gmsh.model.mesh.field.add("Min", n)
    gmsh.model.mesh.field.setNumbers(n, "FieldsList", refinement_fields)
    gmsh.model.mesh.field.setAsBackgroundMesh(n)

    # Turn off default meshing options
    gmsh.model.mesh.MeshSizeFromPoints = 0
    gmsh.model.mesh.MeshSizeFromCurvature = 0
    gmsh.model.mesh.MeshSizeExtendFromBoundary = 0

    occ.synchronize()

    # model.generate_mesh(3)
    gmsh.option.setNumber(
        "General.Terminal", 1 if verbosity else 0
    )  # 1 verbose, 0 otherwise
    gmsh.model.mesh.generate(3)
    gmsh.write(filename)

    # Mesh
    gmsh.finalize()
    return True


if __name__ == "__main__":
    import gdsfactory as gf

    # c = gf.component.Component()
    # waveguide = c << gf.get_component(gf.components.straight_pin(length=10, taper=None))
    # undercut = c << gf.get_component(
    #     gf.components.rectangle(
    #         size=(5.0, 5.0),
    #         layer="UNDERCUT",
    #         centered=True,
    #     )
    # ).move(destination=[4, 0])
    # c.show()

    c = gf.component.Component()
    # waveguide = c << gf.get_component(gf.components.straight_pin(length=5, taper=None))
    ring = c << gf.get_component(gf.components.ring_crow)
    # bend = c << gf.get_component(gf.components.spiral_double(cross_section="rib"))
    # c = gf.components.spiral_racetrack(cross_section='rib')
    c.plot()
    # undercut = c << gf.get_component(
    #     gf.components.rectangle(
    #         size=(5.0, 5.0),
    #         layer="UNDERCUT",
    #         centered=True,
    #     )
    # ).move(destination=[4, 0])
    c.show()

    from gdsfactory.pdk import get_layer_stack

    filtered_layerstack = LayerStack(
        layers={
            k: get_layer_stack().layers[k]
            for k in (
                # "slab90",
                "core",
                # "via_contact",
                # "undercut",
                # "box",
                # "substrate",
                # "clad",
                # "metal1",
            )
        }
    )

    # filtered_layerstack.layers["via_contact"].info["mesh_order"] = 4
    # filtered_layerstack.layers["clad"].info["mesh_order"] = 5

    resolutions = {
        "core": {"resolution": 0.1},
        # "slab90": {"resolution": 0.4},
        # "via_contact": {"resolution": 0.4},
    }
    geometry = xyz_mesh(
        component=c,
        layerstack=filtered_layerstack,
        resolutions=resolutions,
        filename="mesh.msh",
        verbosity=False,
    )
    # print(geometry)

    # import gmsh

    # gmsh.write("mesh.msh")
    # gmsh.clear()
    # geometry.__exit__()

    # import meshio

    # mesh_from_file = meshio.read("mesh.msh")

    # def create_mesh(mesh, cell_type, prune_z=False):
    #     cells = mesh.get_cells_type(cell_type)
    #     cell_data = mesh.get_cell_data("gmsh:physical", cell_type)
    #     points = mesh.points
    #     return meshio.Mesh(
    #         points=points,
    #         cells={cell_type: cells},
    #         cell_data={"name_to_read": [cell_data]},
    #     )

    # line_mesh = create_mesh(mesh_from_file, "line", prune_z=True)
    # meshio.write("facet_mesh.xdmf", line_mesh)

    # triangle_mesh = create_mesh(mesh_from_file, "triangle", prune_z=True)
    # meshio.write("mesh.xdmf", triangle_mesh)

    # # for layer, polygons in heaters.get_polygons(by_spec=True).items():
    # #     print(layer, polygons)



from collections import defaultdict
from functools import partial
from typing import Dict, List, NamedTuple, Union

import bokeh.events as be
import numpy as np
import pandas as pd
import yaml
from bokeh import io as bio
from bokeh import models as bm
from bokeh import plotting as bp
from natsort import natsorted

import gdsfactory as gf
from gdsfactory.picmodel import PicYamlConfiguration, Placement, SchematicConfiguration

data = {
    "srcs": defaultdict(lambda: defaultdict(lambda: [])),
    "dss": {},
}

COLORS_BY_PORT_TYPE = {
    "optical": "#0000ff",
    "electrical": "#00ff00",
    "placement": "white",
    None: "gray",  # default
}


def save_netlist(netlist, filename):
    with open(filename, mode="w") as f:
        d = netlist.dict(exclude_none=True)
        if "placements" in d:
            placements_dict = d["placements"]
        elif "schematic_placements" in d:
            placements_dict = d["schematic_placements"]
        else:
            raise ValueError("No placements attribute found in netlist")
        pkeys = list(placements_dict.keys())
        p: dict = placements_dict
        for pk in pkeys:
            pv = p[pk]
            if pv:
                for kk in ["x", "y"]:
                    if kk in pv:
                        try:
                            pv[kk] = float(pv[kk])
                        except Exception:
                            pass
            else:
                p.pop(pk)
        yaml.dump(d, f, sort_keys=False, default_flow_style=None)


class Rect(NamedTuple):
    tag: str
    x: float
    y: float
    w: float
    h: float
    c: str


class LayerPolygons(NamedTuple):
    tag: str
    xs: List[List[List[float]]]
    ys: List[List[List[float]]]
    c: str
    alpha: float


class LineSegment(NamedTuple):
    tag: str
    x0: float
    y0: float
    x1: float
    y1: float
    name: str


def _enlarge_limits(ax, x, y, w=0.0, h=0.0):
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    x_min, x_max = xlim if xlim != (0.0, 1.0) else (np.inf, -np.inf)
    y_min, y_max = ylim if ylim != (0.0, 1.0) else (np.inf, -np.inf)
    x_min = min(x_min, x)
    x_max = max(x + w, x_max)
    y_min = min(y_min, y)
    y_max = max(y + h, y_max)
    if x_max == x_min:
        x_max += 1.0
    if y_max == y_min:
        y_max += 1.0
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)


def _get_sources(objs):
    srcs = defaultdict(lambda: defaultdict(lambda: []))
    for obj in objs:
        if isinstance(obj, LineSegment):
            src = srcs["MultiLine"]
            src["tag"].append(obj.tag)
            src["x"].append(np.array([obj.x0, obj.x1]))
            src["y"].append(np.array([obj.y0, obj.y1]))
            src["line_color"].append("#000000")
            src["name"].append(obj.name)
        elif isinstance(obj, Rect):
            src = srcs["Rect"]
            src["tag"].append(obj.tag)
            src["x"].append(obj.x + obj.w / 2)
            src["y"].append(obj.y + obj.h / 2)
            src["width"].append(obj.w)
            src["height"].append(obj.h)
            src["fill_color"].append(obj.c)
            src["fill_alpha"].append(0.1)
        elif isinstance(obj, gf.Port):
            src = srcs["Port"]
            src["tag"].append(obj.tag)
            src["x"].append(obj.x)
            src["y"].append(obj.y)
            # src["width"].append(obj.w)
            # src["height"].append(obj.h)
            color = COLORS_BY_PORT_TYPE.get(obj.port_type, COLORS_BY_PORT_TYPE[None])
            src["fill_color"].append(color)
            src["fill_alpha"].append(0.5)
        elif isinstance(obj, LayerPolygons):
            src = srcs["Polygons"]
            src["tag"].append(obj.tag)
            src["xs"].append(obj.xs)
            src["ys"].append(obj.ys)
            src["fill_color"].append(obj.c)
            src["fill_alpha"].append(obj.alpha)
    return srcs


def _get_column_data_sources(srcs):
    _srcs = {}
    for k, src in srcs.items():
        ds = bm.ColumnDataSource(dict(src.items()))
        _srcs[k] = ds
    return _srcs


def viz_bk(
    netlist: Union[SchematicConfiguration, PicYamlConfiguration],
    instances,
    netlist_filename,
    fig=None,
    **kwargs,
):
    global data
    if fig is None:
        fig = bp.figure()

    if isinstance(netlist, (PicYamlConfiguration, SchematicConfiguration)):
        objs = viz_netlist(netlist, instances, **kwargs)
    else:
        objs = netlist
        netlist = None

    if not isinstance(objs, list):
        raise ValueError("viz_bk can only visualize a list of objects.")

    # data['srcs'] = _viz_bk_srcs(objs)

    srcs = _get_sources(objs)
    dss = data["dss"] = _get_column_data_sources(srcs)
    netlist = data["netlist"]

    def cb_rect_on_change_data(attr, old, new):
        tags = np.array(old["tag"], dtype=object)
        xy_old = np.stack([old["x"], old["y"]], 1)
        xy_new = np.stack([new["x"], new["y"]], 1)
        if xy_old.shape != xy_new.shape:
            dss["Rect"].data.__dict__.update(old)
            return
        dxs, dys = (xy_new - xy_old).T  # type: ignore
        idxs = np.where(dxs**2 + dys**2 > 1.0)[0]
        tags = tags[idxs]
        dxs = dxs[idxs]
        dys = dys[idxs]

        for tag, dx, dy in zip(tags, dxs, dys):  # loop over all displaced rectangles
            if netlist is not None:
                if tag not in netlist.placements:
                    netlist.placements[tag] = Placement(x=0, y=0, dx=0, dy=0)
                dx_, dy_ = netlist.placements[tag].dx, netlist.placements[tag].dy
                dx_ = 0.0 if dx_ is None else dx_
                dy_ = 0.0 if dy_ is None else dy_
                netlist.placements[tag].dx = float(dx_ + dx)
                netlist.placements[tag].dy = float(dy_ + dy)
            for k, v in dss.items():
                if k == "Rect":
                    continue
                elif k == "MultiLine":
                    data = dict(v.data)
                    for i, tag_ in enumerate(data["tag"]):
                        if "," in tag_:
                            tag1, tag2 = tag_.split(",")
                            if tag == tag1:
                                data["x"][i][0] += dx
                                data["y"][i][0] += dy
                            elif tag == tag2:
                                data["x"][i][-1] += dx
                                data["y"][i][-1] += dy
                            else:
                                continue
                        else:
                            if tag_ == tag:
                                data["x"][i][:] += dx
                                data["y"][i][:] += dy
                            else:
                                continue
                    v.data = data
                elif k == "Polygons":
                    data = dict(v.data)
                    for i, tag_ in enumerate(data["tag"]):
                        if tag_ == tag:
                            for i_poly in range(len(data["xs"][i])):
                                for i_boundary in range(len(data["xs"][i][i_poly])):
                                    data["xs"][i][i_poly][i_boundary] += dx
                                    data["ys"][i][i_poly][i_boundary] += dy
                    v.data = data
                elif k == "Port":
                    data = dict(v.data)
                    for i, tag_ in enumerate(data["tag"]):
                        if tag_ == tag:
                            data["x"][i] += dx
                            data["y"][i] += dy
                        else:
                            continue
                    v.data = data
        save_netlist(netlist, netlist_filename)

    def cb_rect_selected_on_change_indices(attr, old, new):
        if len(new) > 1:
            data["dss"]["Rect"].selected.indices = [new[0]]

    def cp_double_tap(event):
        # only works on 'hierarchical netlists...'
        if netlist is None:
            return
        df = pd.DataFrame(data["dss"]["Rect"].data)
        mask = np.ones_like(df.x, dtype=bool)
        mask &= df.x - df.width / 2 < event.x
        mask &= event.x < df.x + df.width / 2
        mask &= df.y - df.height / 2 < event.y
        mask &= event.y < df.y + df.width / 2
        df = df[mask]

        tags = df.tag.values
        if tags.shape[0] != 1:
            return

        tag = tags[0]
        if tag in netlist.placements:
            cur_rotation = netlist.placements[tag].rotation or 0
            netlist.placements[tag].rotation = (cur_rotation + 90) % 360
        else:
            return

        update_schematic_plot(schematic=netlist, instances=instances)

    data["dss"]["Rect"].on_change("data", cb_rect_on_change_data)
    data["dss"]["Rect"].selected.on_change(
        "indices", cb_rect_selected_on_change_indices
    )
    fig.on_event(be.DoubleTap, cp_double_tap)

    fig.add_glyph(
        data["dss"]["Rect"],
        bm.Rect(
            x="x",
            y="y",
            width="width",
            height="height",
            fill_color="fill_color",
            fill_alpha="fill_alpha",
        ),
        name="instances",
    )
    if "Polygons" in data["dss"]:
        fig.add_glyph(
            data["dss"]["Polygons"],
            bm.MultiPolygons(
                xs="xs", ys="ys", fill_color="fill_color", fill_alpha="fill_alpha"
            ),
        )
    if "MultiLine" in data["dss"]:
        fig.add_glyph(
            data["dss"]["MultiLine"],
            bm.MultiLine(xs="x", ys="y"),
            name="nets",
        )  # , line_color="line_color"))
    fig.add_glyph(
        data["dss"]["Port"], glyph=bm.Circle(x="x", y="y", fill_color="fill_color")
    )
    del fig.tools[:]
    draw_tool = bm.PointDrawTool(
        renderers=[r for r in fig.renderers if isinstance(r.glyph, bm.Rect)],
        empty_value="black",
    )
    hover_tool = bm.HoverTool(
        names=["instances"],
        tooltips=[("Instance", "@tag")],
    )
    hover_tool_nets = bm.HoverTool(
        names=["nets"],
        tooltips=[("Net", "@name")],
        show_arrow=True,
        line_policy="interp",
    )
    # pan_tool = bm.PanTool()
    tap_tool = bm.TapTool()
    zoom = bm.WheelZoomTool()
    fig.add_tools(draw_tool, hover_tool, hover_tool_nets, tap_tool, zoom)
    fig.toolbar.active_scroll = zoom
    fig.toolbar.active_tap = tap_tool
    fig.toolbar.active_drag = draw_tool
    fig.toolbar.logo = None
    fig.xaxis.major_label_text_font_size = "0pt"
    fig.yaxis.major_label_text_font_size = "0pt"
    fig.match_aspect = True

    def bkapp(doc):
        doc.add_root(fig)
        data["doc"] = doc

    return bkapp


def get_ports(component):
    comp = component
    return natsorted(comp.ports.keys())


def is_output_port(port):
    if "," in port:
        return is_output_port(port.split(",")[-1])
    return port.startswith("out")


def is_input_port(port):
    return not is_output_port(port)


def get_input_ports(component):
    ports = get_ports(component)
    return [p for p in ports if is_input_port(p)]


def get_output_ports(component):
    ports = get_ports(component)
    return [p for p in ports if is_output_port(p)]


def ports_ys(ports, instance_height):
    h = instance_height
    if len(ports) < 1:
        return [h / 2]
    ys = np.linspace(0, h, len(ports) + 1)
    dy = ys[1] - ys[0]
    return ys[1:] - dy / 2


def viz_instance(
    netlist: Union[PicYamlConfiguration, SchematicConfiguration],
    instance_name,
    component,
    instance_size,
):
    # inst_spec = netlist.instances[instance_name].dict()
    inst_ref = component.named_references[instance_name]
    bbox = inst_ref.bbox
    w = bbox[1][0] - bbox[0][0]
    h = bbox[1][1] - bbox[0][1]
    x0 = bbox[0][0]
    y0 = bbox[0][1]
    # pl = w / 10
    # input_ports = get_input_ports(component)
    # output_ports = get_output_ports(component)
    # y_inputs = ports_ys(input_ports, h)
    # y_outputs = ports_ys(output_ports, h)
    # x, y = get_placements(netlist).get(instance_name, (0, 0))
    x, y = x0, y0
    polys_by_layer = inst_ref.get_polygons(by_spec=True, as_array=False)
    layer_polys = []
    layer_views = gf.pdk.get_layer_views()

    for layer, polys in polys_by_layer.items():
        if layer not in layer_views.get_layer_tuples():
            print(f"layer {layer} not found")
            continue
        lv = layer_views.get_from_tuple(layer)
        if lv:
            xs = [[p.points[:, 0]] for p in polys]
            ys = [[p.points[:, 1]] for p in polys]
            lp = LayerPolygons(
                tag=instance_name,
                xs=xs,
                ys=ys,
                c=lv.get_color_dict()["fill_color"],
                alpha=lv.get_alpha(),
            )
            layer_polys.append(lp)

    ports: List[gf.Port] = inst_ref.ports.values()
    ports = [p.copy() for p in ports]
    for p in ports:
        # p.move((x, y))
        p.tag = instance_name
    c = "#000000"

    r = Rect(tag=instance_name, x=x, y=y, w=w, h=h, c=c)
    return [r, *ports] + layer_polys


def split_port(port, netlist):
    if "," not in port:
        port = netlist.ports[port]
    *instance_name, port = port.split(",")
    return ",".join(instance_name), port


def viz_connection(netlist, p_in, p_out, instance_size, point1, point2):
    x1, y1 = point1
    x2, y2 = point2
    tag = f"{p_in.split(',')[0]},{p_out.split(',')[0]}"
    name = f"{p_in} ➔ {p_out}"
    line = LineSegment(tag, x1, y1, x2, y2, name=name)
    return [line]


def viz_netlist(netlist, instances, instance_size=20):
    schematic_dict = netlist.dict()
    schematic_as_layout = {
        "instances": schematic_dict["instances"],
        "placements": schematic_dict["schematic_placements"],
    }
    schematic_component = gf.read.from_yaml(schematic_as_layout, mode="schematic")

    els = []
    port_coords = {}
    for instance_name in netlist.instances:
        els += viz_instance(netlist, instance_name, schematic_component, instance_size)
        for el in els:
            if isinstance(el, gf.Port):
                port_name = f"{instance_name},{el.name}"
                port_coords[port_name] = el.center

    for net in netlist.nets:
        p_in, p_out = net
        point1 = port_coords[p_in]
        point2 = port_coords[p_out]
        els += viz_connection(netlist, p_in, p_out, instance_size, point1, point2)
    return els


def show_netlist(schematic: SchematicConfiguration, instances: Dict, netlist_filename):
    global data
    data["netlist"] = schematic
    fig = bp.figure(width=800, height=500)
    app = viz_bk(
        schematic,
        instances=instances,
        fig=fig,
        instance_size=50,
        netlist_filename=netlist_filename,
    )
    bio.show(app)


def update_schematic_plot(
    schematic: SchematicConfiguration, instances: Dict, *args, **kwargs
):
    global data

    if "doc" in data:
        doc = data["doc"]
        doc.add_next_tick_callback(
            partial(
                _update_schematic_plot,
                schematic=schematic,
                instances=instances,
            )
        )


def _update_schematic_plot(
    schematic: SchematicConfiguration, instances: Dict, *args, **kwargs
):
    srcs = _get_sources(viz_netlist(schematic, instances=instances))
    for k in srcs:
        data["dss"][k].data = srcs[k]


def add_instance(name: str, component):
    inst_viz = viz_instance(
        data["netlist"], instance_name=name, component=component, instance_size=0
    )
    srcs = _get_sources([inst_viz])
    for k, src in srcs.items():
        cds: bm.ColumnDataSource = data["dss"][k]
        cds.stream(src)


def get_deltas(netlist):
    return {
        k: {"dx": p.dx or 0, "dy": p.dy or 0} for k, p in netlist.placements.items()
    }


def apply_deltas(netlist, deltas):
    for k, d in deltas.items():
        netlist.placements[k].dx = d["dx"]
        netlist.placements[k].dy = d["dy"]



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.straight import straight


@gf.cell
def coupler_straight(
    length: float = 10.0,
    gap: float = 0.27,
    **kwargs,
) -> Component:
    """Coupler_straight with two parallel straights.

    Args:
        length: of straight.
        gap: between straights.
        kwargs: cross_section settings.
    """
    component = Component()
    straight_component = straight(length=length, **kwargs)

    top = component << straight_component
    bot = component << straight_component

    # bot.ymax = 0
    # top.ymin = gap

    top.movey(straight_component.info["width"] + gap)

    component.add_port("o1", port=bot.ports["o1"])
    component.add_port("o2", port=top.ports["o1"])
    component.add_port("o3", port=bot.ports["o2"])
    component.add_port("o4", port=top.ports["o2"])
    component.auto_rename_ports()
    return component


if __name__ == "__main__":
    c = coupler_straight(length=2)
    c.show(show_ports=True)



from typing import Dict

from pydantic import BaseModel

from gdsfactory.materials import MaterialSpec
from gdsfactory.materials import (
    material_name_to_lumerical as material_name_to_lumerical_default,
)


class SimulationSettingsLumericalFdtd(BaseModel):
    """Lumerical FDTD simulation_settings.

    Parameters:
        background_material: for the background.
        port_margin: on both sides of the port width (um).
        port_height: port height (um).
        port_extension: port extension (um).
        mesh_accuracy: 2 (1: coarse, 2: fine, 3: superfine).
        zmargin: for the FDTD region (um).
        ymargin: for the FDTD region (um).
        xmargin: for the FDTD region (um).
        wavelength_start: 1.2 (um).
        wavelength_stop: 1.6 (um).
        wavelength_points: 500.
        simulation_time: (s) related to max path length
            3e8/2.4*10e-12*1e6 = 1.25mm.
        simulation_temperature: in kelvin (default = 300).
        frequency_dependent_profile: compute mode profiles for each wavelength.
        field_profile_samples: number of wavelengths to compute field profile.
    """

    background_material: str = "sio2"
    port_margin: float = 1.5
    port_extension: float = 5.0
    mesh_accuracy: int = 2
    zmargin: float = 1.0
    ymargin: float = 3.0
    xmargin: float = 3.0
    wavelength_start: float = 1.2
    wavelength_stop: float = 1.6
    wavelength_points: int = 500
    simulation_time: float = 10e-12
    simulation_temperature: float = 300
    frequency_dependent_profile: bool = True
    field_profile_samples: int = 15
    distance_source_to_monitors: float = 0.2
    material_name_to_lumerical: Dict[
        str, MaterialSpec
    ] = material_name_to_lumerical_default

    class Config:
        """pydantic basemodel config."""

        arbitrary_types_allowed = True


SIMULATION_SETTINGS_LUMERICAL_FDTD = SimulationSettingsLumericalFdtd()

if __name__ == "__main__":
    from gdsfactory.serialization import clean_value_json

    d = clean_value_json(SIMULATION_SETTINGS_LUMERICAL_FDTD)



"""Preprocessing involving both the GDS and the LayerStack, or the resulting simulation polygons."""
from __future__ import annotations

from typing import Dict

import numpy as np
from shapely.affinity import scale
from shapely.geometry import MultiPolygon
from shapely.ops import unary_union

from gdsfactory.simulation.gmsh.parse_gds import to_polygons
from gdsfactory.technology import LayerLevel, LayerStack


def bufferize(layerstack: LayerStack):
    """Convert layers without a z_to_bias to an equivalent z_to_bias.

    Arguments:
        layerstack: layerstack to process
    """
    for layer in layerstack.layers.values():
        if layer.z_to_bias is None:
            if layer.sidewall_angle is None:
                layer.z_to_bias = ([0, 1], [0, 0])
            else:
                buffer_magnitude = layer.thickness * np.tan(
                    np.radians(layer.sidewall_angle)
                )
                layer.z_to_bias = (
                    (
                        [0, layer.width_to_z, 1],
                        [
                            1 * buffer_magnitude * layer.width_to_z,
                            0,
                            -1 * buffer_magnitude * (1 - layer.width_to_z),
                        ],
                    )
                    if layer.width_to_z
                    else ([0, 1], [0, -1 * buffer_magnitude])
                )
    return layerstack


def process_buffers(layer_polygons_dict: Dict, layerstack: LayerStack):
    """Break up layers into sub-layers according to z_to_bias.

    Arguments:
        layer_polygons_dict: dict of GDS layernames: shapely polygons
        layerstack: original Layerstack

    Returns:
        extended_layer_polygons_dict: dict of simulation_layername: (gds_layername, next_simulation_layername, this_layer_polygons, next_layer_polygons)
        extended_layerstack: LayerStack of simulation_layername: simulation layers
    """
    extended_layer_polygons_dict = {}
    extended_layerstack_layers = {}

    layerstack = bufferize(layerstack)

    for layername, polygons in layer_polygons_dict.items():
        zs = layerstack.layers[layername].z_to_bias[0]
        width_buffers = layerstack.layers[layername].z_to_bias[1]
        for ind, (z, width_buffer) in enumerate(zip(zs[:-1], width_buffers[:-1])):
            new_zmin = (
                layerstack.layers[layername].zmin
                + layerstack.layers[layername].thickness * z
            )
            new_thickness = (
                layerstack.layers[layername].thickness * zs[ind + 1]
                - layerstack.layers[layername].thickness * z
            )
            extended_layerstack_layers[f"{layername}_{z}"] = LayerLevel(
                layer=layerstack.layers[layername].layer,
                thickness=new_thickness,
                zmin=new_zmin,
                material=layerstack.layers[layername].material,
                info=layerstack.layers[layername].info,
            )
            extended_layer_polygons_dict[f"{layername}_{z}"] = (
                f"{layername}",
                f"{layername}_{zs[ind+1]}",
                polygons.buffer(width_buffer),
                polygons.buffer(width_buffers[ind + 1]),
            )
        extended_layerstack_layers[f"{layername}_{zs[-1]}"] = LayerLevel(
            layer=layerstack.layers[layername].layer,
            thickness=0,
            zmin=layerstack.layers[layername].zmin
            + layerstack.layers[layername].thickness,
            material=layerstack.layers[layername].material,
            info=layerstack.layers[layername].info,
        )

    return extended_layer_polygons_dict, LayerStack(layers=extended_layerstack_layers)


def buffers_to_lists(layer_polygons_dict: Dict, layerstack: LayerStack):
    """Break up polygons on each layer into lists of polygons:z tuples according to z_to_bias.

    Arguments:
        layer_polygons_dict: dict of GDS layernames: shapely polygons
        layerstack: original Layerstack

    Returns:
        extended_layer_polygons_dict: dict of layername: List[(z, polygon_at_z)] for all polygons at z
    """
    extended_layer_polygons_dict = {}

    layerstack = bufferize(layerstack)

    xfactor, yfactor = 1, 1  # buffer_to_scaling(polygon, width_buffer)
    for layername, polygons in layer_polygons_dict.items():
        all_polygons_list = []
        for polygon in polygons.geoms if hasattr(polygons, "geoms") else [polygons]:
            zs = layerstack.layers[layername].z_to_bias[0]
            width_buffers = layerstack.layers[layername].z_to_bias[1]

            polygons_list = [
                (
                    z
                    * (
                        layerstack.layers[layername].zmin
                        + layerstack.layers[layername].thickness
                    )
                    + layerstack.layers[layername].zmin,
                    scale(polygon, xfact=xfactor, yfact=yfactor),
                )
                for z, width_buffer in zip(zs, width_buffers)
            ]
            all_polygons_list.append(polygons_list)
        extended_layer_polygons_dict[layername] = all_polygons_list

    return extended_layer_polygons_dict


def merge_by_material_func(layer_polygons_dict: Dict, layerstack: LayerStack):
    """Merge polygons of layer_polygons_dict whose layerstack keys share the same material in layerstack values.

    Returns new layer_polygons_dict with merged polygons and materials as keys.
    """
    merged_layer_polygons_dict = {}
    for layername, polygons in layer_polygons_dict.items():
        material = layerstack.layers[layername].material
        if material in merged_layer_polygons_dict:
            merged_layer_polygons_dict[material] = unary_union(
                MultiPolygon(
                    to_polygons([merged_layer_polygons_dict[material], polygons])
                )
            )
        else:
            merged_layer_polygons_dict[material] = polygons

    return merged_layer_polygons_dict


def create_2D_surface_interface(
    layer_polygons: MultiPolygon,
    thickness_min: float = 0.0,
    thickness_max: float = 0.01,
    simplify: float = 0.005,
):
    """Create 2D entity at the interface of two layers/materials.

    Arguments:
        layer_polygons: shapely polygons.
        thickness_min: distance to define the interfacial region towards the polygon.
        thickness_max: distance to define the interfacial region away from the polygon.
        simplify: simplification factor for over-parametrized geometries

    Returns:
        shapely interface polygon
    """
    interfaces = layer_polygons.boundary
    interface_surface = layer_polygons.boundary
    left_hand_side = interfaces.buffer(thickness_max, single_sided=True)
    right_hand_side = interfaces.buffer(-thickness_min, single_sided=True)
    interface_surface = left_hand_side.union(right_hand_side)

    return interface_surface.simplify(simplify, preserve_topology=False)



from __future__ import annotations

from typing import Optional

import numpy as np
from numpy import float64

import gdsfactory as gf
from gdsfactory.add_padding import get_padding_points
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bezier import (
    bezier,
    bezier_curve,
    find_min_curv_bezier_control_points,
)
from gdsfactory.components.ellipse import ellipse
from gdsfactory.components.taper import taper
from gdsfactory.geometry.functions import path_length
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, LayerSpec


def snap_to_grid(p: float, grid_per_unit: int = 1000) -> float64:
    """Round."""
    return np.round(p * grid_per_unit) / grid_per_unit


@cell
def crossing_arm(
    r1: float = 3.0,
    r2: float = 1.1,
    w: float = 1.2,
    L: float = 3.4,
    layer_slab: LayerSpec = "SLAB150",
    cross_section: CrossSectionSpec = "strip",
) -> Component:
    """Returns crossing arm.

    Args:
        r1: ellipse radius1.
        r2: ellipse radius2.
        w: width in um.
        L: length in um.
        layer_slab: for the shallow etch.
        cross_section: spec.
    """
    c = Component()

    layer_slab = gf.get_layer(layer_slab)
    c << ellipse(radii=(r1, r2), layer=layer_slab)

    xs = gf.get_cross_section(cross_section)
    width = xs.width
    layer_wg = gf.get_layer(xs.layer)

    a = np.round(L + w / 2, 3)
    h = width / 2

    taper_pts = [
        (-a, h),
        (-w / 2, w / 2),
        (w / 2, w / 2),
        (a, h),
        (a, -h),
        (w / 2, -w / 2),
        (-w / 2, -w / 2),
        (-a, -h),
    ]

    c.add_polygon(taper_pts, layer=layer_wg)
    c.add_port(
        name="o1",
        center=(-a, 0),
        orientation=180,
        width=width,
        layer=layer_wg,
        cross_section=xs,
    )

    c.add_port(
        name="o2",
        center=(a, 0),
        orientation=0,
        width=width,
        layer=layer_wg,
        cross_section=xs,
    )

    return c


@cell
def crossing(
    arm: ComponentSpec = crossing_arm,
    cross_section: CrossSectionSpec = "strip",
) -> Component:
    """Waveguide crossing.

    Args:
        arm: arm spec.
        cross_section: spec.
    """
    x = gf.get_cross_section(cross_section)
    c = Component()
    arm = gf.get_component(arm)
    arm_h = arm.ref()
    arm_v = arm.ref(rotation=90)

    port_id = 0
    for i in [arm_h, arm_v]:
        c.add(i)
        c.absorb(i)
        for p in i.ports.values():
            c.add_port(name=port_id, port=p)
            port_id += 1
    c.auto_rename_ports()

    x.add_bbox_layers(c)

    if x.cladding_layers and x.cladding_offsets:
        padding = []
        for offset in x.cladding_offsets:
            points = get_padding_points(component=c, default=offset)
        for layer, points in zip(x.bbox_layers, padding):
            c.add_polygon(points, layer=layer)

    if x.add_bbox:
        c = x.add_bbox(c)
    if x.add_pins:
        c = x.add_pins(c)
    return c


@cell
def crossing_from_taper(taper=lambda: taper(width2=2.5, length=3.0)) -> Component:
    """Returns Crossing based on a taper.

    The default is a dummy taper.

    Args:
        taper: taper function.
    """
    taper = gf.get_component(taper)

    c = Component()
    for i, a in enumerate([0, 90, 180, 270]):
        _taper = taper.ref(position=(0, 0), port_id="o2", rotation=a)
        c.add(_taper)
        c.add_port(name=i, port=_taper.ports["o1"])
        c.absorb(_taper)

    c.auto_rename_ports()
    return c


@cell
def crossing_etched(
    width: float = 0.5,
    r1: float = 3.0,
    r2: float = 1.1,
    w: float = 1.2,
    L: float = 3.4,
    layer_wg: LayerSpec = "WG",
    layer_slab: LayerSpec = "SLAB150",
) -> Component:
    """Waveguide crossing.

    - The full crossing has to be on WG layer (to start with a 220nm slab)
    - Then we etch the ellipses down to 150nm slabs and we keep linear taper at 220nm.
    What we write is what we etch on this step

    Args:
        width: input waveguides width.
        r1: radii.
        r2: radii.
        w: wide width.
        L: length.
        layer_wg: waveguide layer.
        layer_slab: shallow etch layer.
    """
    layer_wg = gf.get_layer(layer_wg)
    layer_slab = gf.get_layer(layer_slab)

    # Draw the ellipses
    c = Component()
    ellipse1 = c << ellipse(radii=(r1, r2), layer=layer_wg)
    ellipse2 = c << ellipse(radii=(r2, r1), layer=layer_wg)
    c.absorb(ellipse1)
    c.absorb(ellipse2)

    a = L + w / 2
    h = width / 2

    taper_cross_pts = [
        (-a, h),
        (-w / 2, w / 2),
        (-h, a),
        (h, a),
        (w / 2, w / 2),
        (a, h),
        (a, -h),
        (w / 2, -w / 2),
        (h, -a),
        (-h, -a),
        (-w / 2, -w / 2),
        (-a, -h),
    ]

    c.add_polygon(taper_cross_pts, layer=layer_wg)

    # tapers_poly = c.add_polygon(taper_cross_pts, layer=layer_wg)
    # b = a - 0.1  # To make sure we get 4 distinct polygons when doing bool ops
    # tmp_polygon = [(-b, b), (b, b), (b, -b), (-b, -b)]
    # polys_etch = gdstk.fast_boolean([tmp_polygon], tapers_poly, "not", layer=layer_slab)
    # c.add(polys_etch)

    positions = [(a, 0), (0, a), (-a, 0), (0, -a)]
    angles = [0, 90, 180, 270]

    for i, (p, angle) in enumerate(zip(positions, angles)):
        c.add_port(
            name=str(i),
            center=p,
            orientation=angle,
            width=width,
            layer=layer_wg,
        )
    c.auto_rename_ports()
    return c


@cell
def crossing45(
    crossing: ComponentSpec = crossing,
    port_spacing: float = 40.0,
    dx: Optional[float] = None,
    alpha: float = 0.08,
    npoints: int = 101,
    cross_section: CrossSectionSpec = "strip",
) -> Component:
    r"""Returns 45deg crossing with bends.

    Args:
        crossing: crossing function.
        port_spacing: target I/O port spacing.
        dx: target length.
        alpha: optimization parameter. diminish it for tight bends,
          increase it if raises assertion angle errors
        npoints: number of points.


    The 45 Degree crossing CANNOT be kept as an SRef since
    we only allow for multiples of 90Deg rotations in SRef.

    .. code::

        ----   ----
            \ /
             X
            / \
        ---    ----

    """
    crossing = gf.get_component(crossing)

    c = Component()
    x = c << crossing
    x.rotate(45)

    # Add bends
    p_e = x.ports["o3"].center
    p_w = x.ports["o1"].center
    p_n = x.ports["o2"].center
    p_s = x.ports["o4"].center

    # Flatten the crossing - not an SRef anymore
    c.absorb(x)
    dx = dx or port_spacing
    dy = port_spacing / 2

    start_angle = 45
    end_angle = 0
    cpts = find_min_curv_bezier_control_points(
        start_point=p_e,
        end_point=(dx, dy),
        start_angle=start_angle,
        end_angle=end_angle,
        npoints=npoints,
        alpha=alpha,
    )

    bend = bezier(
        control_points=cpts,
        start_angle=start_angle,
        end_angle=end_angle,
        npoints=npoints,
    )

    tol = 1e-2
    assert abs(bend.info["start_angle"] - start_angle) < tol, print(
        f"{bend.info['start_angle']} differs from {start_angle}"
    )
    assert abs(bend.info["end_angle"] - end_angle) < tol, bend.info["end_angle"]

    b_tr = bend.ref(position=p_e, port_id="o1")
    b_tl = bend.ref(position=p_n, port_id="o1", h_mirror=True)
    b_bl = bend.ref(position=p_w, port_id="o1", rotation=180)
    b_br = bend.ref(position=p_s, port_id="o1", v_mirror=True)

    for cmp_ref in [b_tr, b_br, b_tl, b_bl]:
        # cmp_ref = _cmp.ref()
        c.add(cmp_ref)
        c.absorb(cmp_ref)

    c.info["bezier_length"] = bend.info["length"]
    c.info["min_bend_radius"] = b_br.info["min_bend_radius"]
    c.bezier = bend
    c.crossing = crossing

    c.add_port("o1", port=b_bl.ports["o2"])
    c.add_port("o2", port=b_tl.ports["o2"])
    c.add_port("o3", port=b_tr.ports["o2"])
    c.add_port("o4", port=b_br.ports["o2"])
    c.snap_ports_to_grid()

    x = gf.get_cross_section(cross_section)
    if x.add_bbox:
        c = x.add_bbox(c)
    if x.add_pins:
        c = x.add_pins(c)

    return c


@cell
def compensation_path(
    crossing45: ComponentSpec = crossing45,
    direction: str = "top",
    cross_section: CrossSectionSpec = "strip",
) -> Component:
    r"""Returns Component Path with same path length as the crossing.

    with input and output ports having same y coordinates

    Args:
        crossing45: component that we want to match in path length.
            needs to have .info["components"] with bends and crossing.
        direction: the direction in which the bend should go "top" / "bottom".

    .. code::

          ----       ----
              \     /
               \   /
                \ /
                 X
                / \
               /   \
              /     \
          ----       ----

    Compensation path:

    .. code::

             --+--
           _/     \_
        --/         \--


    """
    import scipy.optimize as so

    x = gf.get_cross_section(cross_section)

    # Get total path length taken by the bends
    crossing45 = gf.get_component(crossing45)
    bezier_length = crossing45.info["bezier_length"]
    length = 2 * bezier_length

    # Find a bezier S-bend with half this length, but with a fixed length
    # governed by the crossing45 X-distance (west to east ports) and
    # the crossing x_distance

    target_bend_length = length / 2

    def get_x_span(cmp):
        return cmp.ports["o3"].x - cmp.ports["o1"].x

    x_span_crossing45 = get_x_span(crossing45)
    x_span_crossing = get_x_span(crossing45.crossing)

    # x span allowed for the bend
    x0 = (x_span_crossing45 - x_span_crossing) / 2

    def get_control_pts(x, y):
        return [(0, 0), (x0 / 2, 0), (x0 / 2, y), (x0, y)]

    def f(y):
        control_points = get_control_pts(x0, y)
        t = np.linspace(0, 1, 51)
        path_points = bezier_curve(t, control_points)
        return path_length(path_points) - target_bend_length

    # the path length of the s-bend between two ports p0 and p1 is :
    # - larger than the euclidean distance L2(p0, p1)
    # - smaller than the manhattan distance DL(p0, p1)
    #
    # This gives the bounds for the brentq root finding

    ya = target_bend_length - x0
    yb = np.sqrt(target_bend_length**2 - x0**2)

    solution = so.root_scalar(f, bracket=[ya, yb], method="brentq")

    y_bend = solution.root
    y_bend = snap_to_grid(y_bend)

    v_mirror = direction != "top"
    sbend = bezier(control_points=get_control_pts(x0, y_bend))

    c = Component()
    crossing0 = c << crossing45.crossing

    sbend_left = sbend.ref(
        position=crossing0.ports["o1"], port_id="o2", v_mirror=v_mirror
    )
    sbend_right = sbend.ref(
        position=crossing0.ports["o3"], port_id="o2", h_mirror=True, v_mirror=v_mirror
    )

    c.add(sbend_left)
    c.add(sbend_right)

    c.add_port("o1", port=sbend_left.ports["o1"])
    c.add_port("o2", port=sbend_right.ports["o1"])

    c.info["min_bend_radius"] = sbend.info["min_bend_radius"]
    c.info["sbend"] = sbend.info

    if x.cladding_layers and x.cladding_offsets:
        padding = []
        for offset in x.cladding_offsets:
            points = get_padding_points(component=c, default=offset)
        for layer, points in zip(x.bbox_layers, padding):
            c.add_polygon(points, layer=layer)

    if x.add_bbox:
        c = x.add_bbox(c)
    if x.add_pins:
        c = x.add_pins(c)
    return c


def _demo() -> None:
    """Plot curvature of bends."""
    from matplotlib import pyplot as plt

    c = crossing45(port_spacing=20.0, dx=15)
    c2 = compensation_path(crossing45=c)
    print(c.info["min_bend_radius"])
    print(c2.info["min_bend_radius"])

    component = Component(name="top_lvl")
    component.add(c.ref(port_id="o1"))
    component.add(c2.ref(port_id="o1", position=(0, 10)))

    bend_info1 = c.info["components"]["bezier_bend"].info
    bend_info2 = c2.info["components"]["sbend"].info

    DL = bend_info1["length"]
    L2 = bend_info1["length"]
    plt.plot(bend_info1["t"][1:-1] * DL, abs(bend_info1["curvature"]))
    plt.plot(bend_info2["t"][1:-1] * L2, abs(bend_info2["curvature"]))
    plt.xlabel("bend length (um)")
    plt.ylabel("curvature (um^-1)")
    component.show()
    plt.show()


if __name__ == "__main__":
    # c = crossing45()
    c = compensation_path()
    # c = crossing(
    #     cross_section=dict(
    #         cross_section="strip",
    #         settings=dict(cladding_offsets=[0], cladding_layers=[(3, 0)]),
    #     )
    # )
    # print(c.ports["E1"].y - c.ports['o2'].y)
    # print(c.get_ports_array())
    # _demo()
    # c = crossing_from_taper()
    # c.pprint()
    # c = crossing_etched()
    # c = compensation_path()
    # c = crossing45(port_spacing=40)
    c.show(show_ports=True)



"""based on phidl.geometry."""
from __future__ import annotations

from typing import Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.components.rectangle import rectangle
from gdsfactory.components.text import text
from gdsfactory.typings import Anchor, ComponentSpec, LayerSpec

big_square = gf.partial(rectangle, size=(1300, 2600))


@gf.cell
def die_bbox(
    component: ComponentSpec = big_square,
    street_width: float = 100.0,
    street_length: Optional[float] = None,
    die_name: Optional[str] = None,
    text_size: float = 100.0,
    text_anchor: Anchor = "sw",
    layer: LayerSpec = "M3",
    padding: float = 10.0,
) -> gf.Component:
    """Returns component with boundary box frame around it.

    Perfect for defining the boundary of the chip/die
    it can also add a label with the name of the die.
    similar to die and bbox.

    Args:
        component: to frame.
        street_width: Width of the boundary box.
        street_length: length of the boundary box.
        die_name: Label text.
        text_size: Label text size.
        text_anchor: {'nw', 'nc', 'ne', 'sw', 'sc', 'se'} text location.
        layer: Specific layer(s) to put polygon geometry on.
        padding: adds padding.
    """
    D = gf.Component()
    component = gf.get_component(component)

    D.copy_child_info(component)
    cref = D.add_ref(component)
    cref.x = 0
    cref.y = 0
    size = cref.size
    sx, sy = size[0] / 2, size[1] / 2

    sx += street_width + padding
    sy += street_width + padding

    street_length = street_length or max([sx, sy])

    xpts = np.array(
        [
            sx,
            sx,
            sx - street_width,
            sx - street_width,
            sx - street_length,
            sx - street_length,
        ]
    )
    ypts = np.array(
        [
            sy,
            sy - street_length,
            sy - street_length,
            sy - street_width,
            sy - street_width,
            sy,
        ]
    )
    D.add_polygon([+xpts, +ypts], layer=layer)
    D.add_polygon([-xpts, +ypts], layer=layer)
    D.add_polygon([+xpts, -ypts], layer=layer)
    D.add_polygon([-xpts, -ypts], layer=layer)

    if die_name:
        t = D.add_ref(text(text=die_name, size=text_size, layer=layer))

        d = street_width + 20
        if text_anchor == "nw":
            t.xmin, t.ymax = [-sx + d, sy - d]
        elif text_anchor == "nc":
            t.x, t.ymax = [0, sy - d]
        elif text_anchor == "ne":
            t.xmax, t.ymax = [sx - d, sy - d]
        if text_anchor == "sw":
            t.xmin, t.ymin = [-sx + d, -sy + d]
        elif text_anchor == "sc":
            t.x, t.ymin = [0, -sy + d]
        elif text_anchor == "se":
            t.xmax, t.ymin = [sx - d, -sy + d]

    return D


if __name__ == "__main__":
    mask = gf.components.array(rows=10, columns=10)
    # c = die_bbox(component=mask, die_name="chip99")
    c = die_bbox()
    # c.show(show_ports=True)



"""GDS regression test. Adapted from lytest.

TODO: adapt it into pytest_regressions
from __future__ import annotations
from pytest_regressions.file_regression import FileRegressionFixture
class GdsRegressionFixture(FileRegressionFixture):
    def check(self,
        contents,
        extension=".gds",
        basename=None,
        fullpath=None,
        binary=False,
        obtained_filename=None,
        check_fn=None,
            ):
        try:
            difftest(c)
"""
import filecmp
import os
import pathlib
import shutil
from typing import Optional

from gdsfactory.component import Component
from gdsfactory.config import PATH, logger
from gdsfactory.gdsdiff.gdsdiff import gdsdiff


class GeometryDifference(Exception):
    pass


def run_xor(file1, file2, tolerance: int = 1, verbose: bool = False) -> None:
    """Returns nothing.

    Raises a GeometryDifference if there are differences detected.

    Args:
        file1: ref gdspath.
        file2: run gdspath.
        tolerance: in nm.
        verbose: prints output.
    """
    import klayout.db as kdb

    l1 = kdb.Layout()
    l1.read(file1)

    l2 = kdb.Layout()
    l2.read(file2)

    # Check that same set of layers are present
    layer_pairs = []
    for ll1 in l1.layer_indices():
        li1 = l1.get_info(ll1)
        ll2 = l2.find_layer(l1.get_info(ll1))
        if ll2 is None:
            raise GeometryDifference(
                f"Layer {li1} of layout {file1!r} not present in layout {file2!r}."
            )

        layer_pairs.append((ll1, ll2))

    for ll2 in l2.layer_indices():
        li2 = l2.get_info(ll2)
        ll1 = l1.find_layer(l2.get_info(ll2))
        if ll1 is None:
            raise GeometryDifference(
                f"Layer {li2} of layout {file2!r} not present in layout {file1!r}."
            )

    # Check that topcells are the same
    tc1_names = [tc.name for tc in l1.top_cells()]
    tc2_names = [tc.name for tc in l2.top_cells()]
    tc1_names.sort()
    tc2_names.sort()
    if tc1_names != tc2_names:
        raise GeometryDifference(
            f"Missing topcell on one of the layouts, or name differs:\n{tc1_names!r}\n{tc2_names!r}"
        )
    topcell_pairs = [(l1.cell(tc1_n), l2.cell(tc1_n)) for tc1_n in tc1_names]
    # Check that dbu are the same
    if (l1.dbu - l2.dbu) > 1e-6:
        raise GeometryDifference(
            f"Database unit of layout {file1!r} ({l1.dbu}) differs from that of layout {file2!r} ({l2.dbu})."
        )

    # Run the difftool
    diff = False
    for tc1, tc2 in topcell_pairs:
        for ll1, ll2 in layer_pairs:
            r1 = kdb.Region(tc1.begin_shapes_rec(ll1))
            r2 = kdb.Region(tc2.begin_shapes_rec(ll2))

            rxor = r1 ^ r2

            if tolerance > 0:
                rxor.size(-tolerance)

            if not rxor.is_empty():
                diff = True
                if verbose:
                    print(
                        f"{rxor.size()} differences found in {tc1.name!r} on layer {l1.get_info(ll1)}."
                    )

            elif verbose:
                print(
                    f"No differences found in {tc1.name!r} on layer {l1.get_info(ll1)}."
                )

    if diff:
        fn_abgd = []
        for fn in [file1, file2]:
            head, tail = os.path.split(fn)
            abgd = os.path.join(os.path.basename(head), tail)
            fn_abgd.append(abgd)
        raise GeometryDifference(
            "Differences found between layouts {} and {}".format(*fn_abgd)
        )


def difftest(
    component: Component,
    test_name: Optional[str] = None,
    dirpath: pathlib.Path = PATH.gdsdiff,
) -> None:
    """Avoids GDS regressions tests on the GeometryDifference.

    If files are the same it returns None. If files are different runs XOR
    between new component and the GDS reference stored in dirpath and
    raises GeometryDifference if there are differences and show differences in KLayout.

    If it runs for the fist time it just stores the GDS reference.

    Args:
        component: to test if it has changed.
        test_name: used to store the GDS file.
        dirpath: defaults to cwd refers to where the test is being invoked.
    """
    # containers function_name is different from component.name
    # we store the container with a different name from original component
    test_name = test_name or (
        f"{component.function_name}_{component.name}"
        if hasattr(component, "function_name")
        and component.name != component.function_name
        else f"{component.name}"
    )
    filename = f"{test_name}.gds"
    ref_file = dirpath / "gds_ref" / filename
    run_file = dirpath / "gds_run" / filename
    diff_file = dirpath / "gds_diff" / filename

    component.write_gds(gdspath=run_file)

    if not ref_file.exists():
        component.write_gds(gdspath=ref_file)
        raise AssertionError(
            f"Reference GDS file for {test_name!r} not found. Writing to {ref_file!r}"
        )

    if filecmp.cmp(ref_file, run_file, shallow=False):
        return

    try:
        run_xor(str(ref_file), str(run_file), tolerance=1, verbose=False)
    except GeometryDifference as error:
        logger.error(error)
        diff = gdsdiff(ref_file, run_file, name=test_name, xor=False)
        diff.write_gds(diff_file)
        diff.show(show_ports=False)
        print(
            f"\ngds_run {filename!r} changed from gds_ref {str(ref_file)!r}\n"
            "You can check the differences in Klayout GUI or run XOR with\n"
            f"gf gds diff --xor {ref_file} {run_file}\n"
        )

        try:
            val = input(
                "Save current GDS as new reference (Y) or show differences (d)? [Y/n/d]"
            )
            if val.upper().startswith("N"):
                raise
            xor = val.upper().startswith("D")
            if xor:
                diff = gdsdiff(ref_file, run_file, name=test_name, xor=xor)
                diff.write_gds(diff_file)
                diff.show(show_ports=False)

                val = input("Save current GDS as the new reference (Y)? [Y/n]")
                if val.upper().startswith("N"):
                    raise

            logger.info(f"deleting file {str(ref_file)!r}")
            ref_file.unlink()
            shutil.copy(run_file, ref_file)
            raise
        except OSError as exc:
            raise GeometryDifference(
                "\n"
                f"{filename!r} changed from reference {str(ref_file)!r}\n"
                "To step over each error you can run `pytest -s`\n"
                "So you can check the differences in Klayout GUI\n"
            ) from exc


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.components.straight()
    difftest(c)
    # test_component(c, None, None)



"""Plot gdsfactory Components in matplotlib.

based on phidl.quickplotter.
"""

from __future__ import annotations

import sys
from typing import Optional

import numpy as np

from gdsfactory.component import Component
from gdsfactory.component_layout import Polygon, _rotate_points
from gdsfactory.component_reference import ComponentReference

_SUBPORT_RGB = (0, 120, 120)
_PORT_RGB = (190, 0, 0)


try:
    from PyQt5 import QtCore, QtGui
    from PyQt5.QtCore import (
        QCoreApplication,
        QLineF,
        QPoint,
        QPointF,
        QRect,
        QRectF,
        QSize,
        QSizeF,
        Qt,
    )
    from PyQt5.QtGui import QColor, QPen, QPolygonF
    from PyQt5.QtWidgets import (
        QApplication,
        QGraphicsItem,
        QGraphicsScene,
        QGraphicsView,
        QLabel,
        QMainWindow,
        QMessageBox,
        QRubberBand,
    )

    PORT_COLOR = QColor(*_PORT_RGB)
    SUBPORT_COLOR = QColor(*_SUBPORT_RGB)
    OUTLINE_PEN = QColor(200, 200, 200)
    qt_imported = True
except ImportError:
    QMainWindow = object
    QGraphicsView = object
    qt_imported = False

_quickplot_options = {
    "show_ports": True,
    "show_subports": True,
    "label_aliases": False,
    "new_window": False,
    "blocking": False,
    "zoom_factor": 1.4,
    "interactive_zoom": None,
    "fontsize": 14,
}


def _zoom_factory(axis, scale_factor=1.4):
    """Returns zooming functionality to axis.

    From https://gist.github.com/tacaswell/3144287

    """

    def zoom_fun(event, ax, scale):
        """Zoom when scrolling."""
        if event.inaxes == axis:
            scale_factor = np.power(scale, -event.step)
            xdata = event.xdata
            ydata = event.ydata
            x_left = xdata - ax.get_xlim()[0]
            x_right = ax.get_xlim()[1] - xdata
            y_top = ydata - ax.get_ylim()[0]
            y_bottom = ax.get_ylim()[1] - ydata

            ax.set_xlim([xdata - x_left * scale_factor, xdata + x_right * scale_factor])
            ax.set_ylim([ydata - y_top * scale_factor, ydata + y_bottom * scale_factor])
            ax.figure.canvas.draw()
            # Update toolbar so back/forward buttons work
            fig.canvas.toolbar.push_current()

    fig = axis.get_figure()
    fig.canvas.mpl_connect(
        "scroll_event", lambda event: zoom_fun(event, axis, scale_factor)
    )


_qp_objects = {}


def _rectangle_selector_factory(fig, ax):
    from matplotlib.widgets import RectangleSelector

    def line_select_callback(eclick, erelease):
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        left = min(x1, x2)
        right = max(x1, x2)
        bottom = min(y1, y2)
        top = max(y1, y2)
        ax.set_xlim([left, right])
        ax.set_ylim([bottom, top])
        ax.figure.canvas.draw()
        # Update toolbar so back/forward buttons work
        fig.canvas.toolbar.push_current()

    rs = RectangleSelector(
        ax,
        line_select_callback,
        drawtype="box",
        useblit=True,
        button=[1, 3],
        minspanx=5,
        minspany=5,
        spancoords="pixels",
        interactive=False,
    )
    return rs


def set_quickplot_options(
    show_ports: Optional[bool] = None,
    show_subports: Optional[bool] = None,
    label_aliases: Optional[bool] = None,
    new_window: Optional[bool] = None,
    blocking: Optional[bool] = None,
    zoom_factor: Optional[float] = None,
    interactive_zoom: Optional[bool] = None,
    fontsize: Optional[int] = None,
) -> None:
    """Sets plotting options for quickplot().

    Args:
        show_ports: Sets whether ports are drawn.
        show_subports: Sets whether subports (ports that belong to references) are drawn.
        label_aliases: Sets whether aliases are labeled with a text name.
        new_window: If True, each call to quickplot() will generate a separate window.
        blocking: If True, calling quickplot() will pause execution of ("block") the
            remainder of the python code until the quickplot() window is closed.
            If False, the window will be opened and code will continue to run.
        zoom_factor: Sets the scaling factor when zooming the quickplot window with the
            mousewheel/trackpad.
        interactive_zoom: Enables using mousewheel/trackpad to zoom.
        fontsize: for labels.

    """
    if show_ports is not None:
        _quickplot_options["show_ports"] = show_ports
    if show_subports is not None:
        _quickplot_options["show_subports"] = show_subports
    if label_aliases is not None:
        _quickplot_options["label_aliases"] = label_aliases
    if new_window is not None:
        _quickplot_options["new_window"] = new_window
    if blocking is not None:
        _quickplot_options["blocking"] = blocking
    if zoom_factor is not None:
        _quickplot_options["zoom_factor"] = zoom_factor
    if interactive_zoom is not None:
        _quickplot_options["interactive_zoom"] = interactive_zoom
    if fontsize is not None:
        _quickplot_options["fontsize"] = fontsize


def quickplot(items, **kwargs):  # noqa: C901
    """Takes a list of devices/references/polygons or single one of those, and \
    plots them. Use `set_quickplot_options()` to modify the viewer behavior \
    (e.g. displaying ports, creating new windows, etc).

    Args:
        items: object or list of objects to plot.

    Kwargs:
        show_ports: Sets whether ports are drawn.
        show_subports: Sets whether subports (ports that belong to references) are drawn.
        label_aliases: Sets whether aliases are labeled with a text name.
        new_window: If True, each call to quickplot() will generate a separate window.
        blocking: If True, calling quickplot() will pause execution of ("block") the
            remainder of the python code until the quickplot() window is closed.
            If False, the window will be opened and code will continue to run.
        zoom_factor: Sets the scaling factor when zooming the quickplot window with the
            mousewheel/trackpad.
        interactive_zoom: Enables using mousewheel/trackpad to zoom.
        fontsize: for labels.


    Examples
    --------
    >>> import gdsfactory as gf
    >>> R = gf.components.rectangle()
    >>> R.plot()

    >>> E = gf.components.ellipse()
    >>> E.plot()

    """
    from matplotlib import pyplot as plt

    from gdsfactory.path import Path
    from gdsfactory.pdk import get_layer_views
    from gdsfactory.technology import LayerView

    quickplot_options = _quickplot_options.copy()
    quickplot_options.update(**kwargs)

    # Override default options with quickplot_options
    show_ports = quickplot_options["show_ports"]
    show_subports = quickplot_options["show_subports"]
    label_aliases = quickplot_options["label_aliases"]
    new_window = quickplot_options["new_window"]
    blocking = quickplot_options["blocking"]

    if new_window:
        fig, ax = plt.subplots(1)
        ax.autoscale(enable=True, tight=True)
    elif plt.fignum_exists(num="quickplot"):
        fig = plt.figure("quickplot")
        plt.clf()  # Erase figure so toolbar at top works correctly
        ax = fig.add_subplot(111)
    else:
        fig, ax = plt.subplots(num="quickplot")

    ax.axis("equal")
    ax.grid(True, which="both", alpha=0.4)
    ax.axhline(y=0, color="k", alpha=0.2, linewidth=1)
    ax.axvline(x=0, color="k", alpha=0.2, linewidth=1)
    bbox = None

    # Iterate through each Component/ComponentReference/Polygon
    if not isinstance(items, list):
        items = [items]
    LAYER_VIEWS = get_layer_views()
    all_lv_tuples = LAYER_VIEWS.get_layer_tuples()
    for item in items:
        if isinstance(item, (Component, ComponentReference)):
            polygons_spec = item.get_polygons(by_spec=True, depth=None)
            for key in sorted(polygons_spec):
                polygons = polygons_spec[key]
                layer_view = (
                    LAYER_VIEWS.get_from_tuple(key)
                    if key in all_lv_tuples
                    else LayerView(layer=key)
                )
                colors = layer_view.get_color_dict()
                new_bbox = _draw_polygons(
                    polygons,
                    ax,
                    facecolor=colors["fill_color"],
                    edgecolor=colors["frame_color"],
                    alpha=layer_view.get_alpha(),
                )
                bbox = _update_bbox(bbox, new_bbox)
            # If item is a Component or ComponentReference, draw ports
            if isinstance(item, (Component, ComponentReference)) and show_ports is True:
                for port in item.ports.values():
                    if (
                        (port.width is None)
                        or (port.width == 0)
                        or port.orientation is None
                    ):
                        new_bbox = _draw_port_as_point(ax, port)
                    else:
                        new_bbox = _draw_port(ax, port, is_subport=False, color="r")
                    bbox = _update_bbox(bbox, new_bbox)
            if isinstance(item, Component) and show_subports is True:
                for sd in item.references:
                    for port in sd.ports.values():
                        new_bbox = _draw_port(
                            ax,
                            port,
                            is_subport=True,
                            color=np.array(_SUBPORT_RGB) / 255,
                        )
                        bbox = _update_bbox(bbox, new_bbox)
            if isinstance(item, Component) and label_aliases is True:
                for name, ref in item.named_references.items():
                    ax.text(
                        ref.x,
                        ref.y,
                        str(name),
                        style="italic",
                        color="blue",
                        weight="bold",
                        ha="center",
                        fontsize=quickplot_options["fontsize"],
                    )
        elif isinstance(item, Polygon):
            layer_tuple = (item.layer, item.datatype)
            layer_view = (
                LAYER_VIEWS.get_from_tuple(layer_tuple)
                if layer_tuple in all_lv_tuples
                else LayerView(layer=layer_tuple)
            )
            colors = layer_view.get_color_dict()
            new_bbox = _draw_polygons(
                item.points,
                ax,
                facecolor=colors["fill_color"],
                edgecolor=colors["frame_color"],
                alpha=layer_view.get_alpha(),
            )
            bbox = _update_bbox(bbox, new_bbox)
        elif isinstance(item, Path):
            points = item.points
            new_bbox = _draw_line(
                x=points[:, 0],
                y=points[:, 1],
                ax=ax,
                linestyle="--",
                linewidth=2,
                color="b",
            )
            bbox = _update_bbox(bbox, new_bbox)

    if bbox is None:
        bbox = [-1, -1, 1, 1]
    xmargin = (bbox[2] - bbox[0]) * 0.1 + 1e-9
    ymargin = (bbox[3] - bbox[1]) * 0.1 + 1e-9
    ax.set_xlim([bbox[0] - xmargin, bbox[2] + xmargin])
    ax.set_ylim([bbox[1] - ymargin, bbox[3] + ymargin])

    # When using inline Jupyter notebooks, this may fail so allow it to fail gracefully
    try:
        if _use_interactive_zoom():
            _zoom_factory(ax, scale_factor=quickplot_options["zoom_factor"])
        # Need to hang on to RectangleSelector so it doesn't get garbage collected
        _qp_objects["rectangle_selector"] = _rectangle_selector_factory(fig, ax)
        # Update matplotlib toolbar so the Home button works
        fig.canvas.toolbar.update()
        fig.canvas.toolbar.push_current()
    except Exception:
        pass

    plt.draw()
    plt.show(block=blocking)
    return fig


def _use_interactive_zoom():
    """Checks whether the current matplotlib backend is compatible with \
    interactive zoom."""
    import matplotlib

    if _quickplot_options["interactive_zoom"] is not None:
        return _quickplot_options["interactive_zoom"]
    forbidden_backends = ["nbagg"]
    backend = matplotlib.get_backend().lower()
    return all(fb.lower() not in backend for fb in forbidden_backends)


def _update_bbox(bbox, new_bbox):
    if bbox is None:
        return new_bbox
    if new_bbox[0] < bbox[0]:
        bbox[0] = new_bbox[0]  # xmin
    if new_bbox[1] < bbox[1]:
        bbox[1] = new_bbox[1]  # ymin
    if new_bbox[2] > bbox[2]:
        bbox[2] = new_bbox[2]  # xmin
    if new_bbox[3] > bbox[3]:
        bbox[3] = new_bbox[3]  # ymin
    return bbox


def _draw_polygons(polygons, ax, **kwargs):
    from matplotlib.collections import PolyCollection

    coll = PolyCollection(polygons, **kwargs)
    ax.add_collection(coll)
    stacked_polygons = np.vstack(polygons)
    xmin, ymin = np.min(stacked_polygons, axis=0)
    xmax, ymax = np.max(stacked_polygons, axis=0)
    return [xmin, ymin, xmax, ymax]


def _draw_line(x, y, ax, **kwargs):
    from matplotlib.lines import Line2D

    line = Line2D(x, y, **kwargs)
    ax.add_line(line)
    xmin, ymin = np.min(x), np.min(y)
    xmax, ymax = np.max(x), np.max(y)
    return [xmin, ymin, xmax, ymax]


def _port_marker(port, is_subport):
    angle = port.orientation if port.orientation is not None else 0

    if is_subport:
        arrow_scale = 0.75
        rad = (angle + 45) * np.pi / 180
        pm = +1
    else:
        arrow_scale = 1
        rad = (angle - 45) * np.pi / 180
        pm = -1
    arrow_points = (
        np.array([[0, 0], [10, 0], [6, pm * 4], [6, pm * 2], [0, pm * 2]])
        / 35
        * port.width
        * arrow_scale
    )
    arrow_points += port.center
    arrow_points = _rotate_points(arrow_points, angle=angle, center=port.center)
    text_pos = np.array([np.cos(rad), np.sin(rad)]) * port.width / 3 + port.center

    arrow_points = (
        np.array([[0, 0], [0, 0]]) if port.orientation is None else arrow_points
    )
    return arrow_points, text_pos


def _draw_port(ax, port, is_subport, color):
    xbound, ybound = np.column_stack(port.endpoints)
    # plt.plot(x, y, 'rp', markersize = 12) # Draw port center
    arrow_points, text_pos = _port_marker(port, is_subport)
    xmin, ymin = np.min(np.vstack([arrow_points, port.endpoints]), axis=0)
    xmax, ymax = np.max(np.vstack([arrow_points, port.endpoints]), axis=0)
    ax.plot(xbound, ybound, alpha=0.5, linewidth=3, color=color)  # Draw port edge
    ax.plot(
        arrow_points[:, 0], arrow_points[:, 1], alpha=0.8, linewidth=2, color=color
    )  # Draw port edge
    ax.text(
        text_pos[0],
        text_pos[1],
        port.name,
        horizontalalignment="center",
        verticalalignment="center",
        fontsize=14,
        color=color,
    )
    return [xmin, ymin, xmax, ymax]


def _draw_port_as_point(ax, port, **kwargs):
    from matplotlib import pyplot as plt

    x = port.center[0]
    y = port.center[1]
    plt.plot(x, y, "r+", alpha=0.5, markersize=15, markeredgewidth=2)  # Draw port edge
    bbox = [
        x - port.width / 2,
        y - port.width / 2,
        x + port.width / 2,
        y + port.width / 2,
    ]
    ax.text(port.center[0], port.center[1], port.name, fontsize=14)
    return bbox


class ViewerWindow(QMainWindow):
    def __init__(self):
        """Initialize the object."""
        super().__init__()

        self.setGeometry(QRect(100, 100, 800, 600))
        self.setWindowTitle("PHIDL quickplot")

        # Create "grid size = 40.0" label
        self.gridsize_label = QLabel("ABCDEF", self)
        self.gridsize_label.setFont(QtGui.QFont("SansSerif", 10))
        self.gridsize_label.move(0, 200)
        self.gridsize_label.setAlignment(Qt.AlignLeft)
        self.gridsize_label.setStyleSheet("color: gray")
        self.gridsize_label.setFixedWidth(120)
        self.gridsize_label.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)

        # Create "X=40.001, Y = 70.183" label
        self.position_label = QLabel("ABCDEF", self)
        self.position_label.setFont(QtGui.QFont("SansSerif", 10))
        self.position_label.move(50, 200)
        self.position_label.setAlignment(Qt.AlignRight)
        self.position_label.setStyleSheet("color: gray")
        self.position_label.setFixedWidth(240)
        self.position_label.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)

        # Create "Press ? for help" label
        self.help_label = QLabel("ABCDEF", self)
        self.help_label.setFont(QtGui.QFont("SansSerif", 10))
        self.help_label.move(50, 200)
        self.help_label.setAlignment(Qt.AlignCenter)
        self.help_label.setStyleSheet("color: gray")
        self.help_label.setFixedWidth(200)
        self.help_label.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)

        # Create label useful for debugging
        self.debug_label = QLabel("", self)
        self.debug_label.setFont(QtGui.QFont("SansSerif", 10))
        self.debug_label.move(200, 200)
        self.debug_label.setAlignment(Qt.AlignCenter)
        self.debug_label.setStyleSheet("color: gray")
        self.debug_label.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)

        # Create QGraphicsView
        self.viewer = Viewer(
            gridsize_label=self.gridsize_label,
            position_label=self.position_label,
            help_label=self.help_label,
        )
        self.setCentralWidget(self.viewer)

        # Reorder widgets
        self.gridsize_label.raise_()
        self.position_label.raise_()
        self.debug_label.raise_()
        self.help_label.raise_()
        self.show()


class Viewer(QGraphicsView):
    def __init__(self, gridsize_label, position_label, help_label):
        """Initialize the object."""
        QGraphicsView.__init__(self)

        self.gridsize_label = gridsize_label
        self.position_label = position_label
        self.help_label = help_label

        # Create a QGraphicsScene which this view looks at
        self.scene = QGraphicsScene(self)
        self.scene.setSceneRect(QRectF())
        self.setScene(self.scene)

        # Customize QGraphicsView
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setInteractive(False)
        self.scale(1, -1)  # Flips around the Y axis
        # Use OpenGL http://ralsina.me/stories/BBS53.html
        #        self.setViewport(QtOpenGL.QGLWidget())
        self.rubberBand = QRubberBand(QRubberBand.Rectangle, self)
        self.pen = QPen(QtCore.Qt.black, 0)
        self.portpen = QPen(PORT_COLOR, 3)
        self.portpen.setCosmetic(True)  # Makes constant width
        self.portfont = QtGui.QFont("Arial", pointSize=14)
        self.portfontcolor = PORT_COLOR
        self.subportpen = QPen(SUBPORT_COLOR, 3)
        self.subportpen.setCosmetic(True)  # Makes constant width
        self.subportfont = QtGui.QFont("Arial", pointSize=14)
        self.subportfontcolor = SUBPORT_COLOR

        # Tracking ports

        # Various status variables
        self._mousePressed = None
        self._rb_origin = QPoint()
        self.zoom_factor_total = 1

        # Grid variables
        self.gridpen = QPen(QtCore.Qt.black, 0)
        self.gridpen.setStyle(QtCore.Qt.DotLine)
        self.gridpen.setDashPattern([1, 4])
        self.gridpen.setColor(QtGui.QColor(0, 0, 0, 125))
        #        self.gridpen = QPen(QtCore.Qt.black, 1)
        #        self.gridpen.setCosmetic(True) # Makes constant width
        self.scene_polys = []

        self.initialize()

    def add_polygons(
        self, polygons, fill_color="#A8F22A", frame_color="#A8F22A", alpha=1.0
    ):
        qcolor_fill = QColor()
        qcolor_fill.setNamedColor(fill_color)
        qcolor_fill.setAlphaF(alpha)
        qcolor_frame = QColor()
        qcolor_frame.setNamedColor(frame_color)
        qcolor_frame.setAlphaF(1.0)
        qpen_frame = QPen(qcolor_frame, 0)
        for points in polygons:
            qpoly = QPolygonF([QPointF(p[0], p[1]) for p in points])
            scene_poly = self.scene.addPolygon(qpoly)
            scene_poly.setBrush(qcolor_fill)
            scene_poly.setPen(qpen_frame)

            self.scene_polys.append(scene_poly)
            # Update custom bounding box
            sr = scene_poly.sceneBoundingRect()
            if len(self.scene_polys) == 1:
                self.scene_xmin = sr.left()
                self.scene_xmax = sr.right()
                self.scene_ymin = sr.top()
                self.scene_ymax = sr.bottom()
            else:
                self.scene_xmin = min(self.scene_xmin, sr.left())
                self.scene_xmax = max(self.scene_xmax, sr.right())
                self.scene_ymin = min(self.scene_ymin, sr.top())
                self.scene_ymax = max(self.scene_ymax, sr.bottom())

    def reset_view(self):
        # The SceneRect controls how far you can pan, make it larger than
        # just the bounding box so middle-click panning works
        panning_rect = QRectF(self.scene_bounding_rect)
        panning_rect_center = panning_rect.center()
        panning_rect_size = max(panning_rect.width(), panning_rect.height()) * 3
        panning_rect.setSize(QSizeF(panning_rect_size, panning_rect_size))
        panning_rect.moveCenter(panning_rect_center)
        self.setSceneRect(panning_rect)
        self.fitInView(self.scene_bounding_rect, Qt.KeepAspectRatio)
        self.zoom_view(0.8)

        self.update_grid()

    def add_port(self, port, is_subport=False):
        if (port.width is None) or (port.width == 0):
            x, y = port.center
            cs = 1  # cross size
            pn = QPointF(x, y + cs)
            ps = QPointF(x, y - cs)
            pe = QPointF(x + cs, y)
            pw = QPointF(x - cs, y)
            qline1 = self.scene.addLine(QLineF(pn, ps))
            qline2 = self.scene.addLine(QLineF(pw, pe))
            port_shapes = [qline1, qline2]
        else:
            point1, point2 = port.endpoints
            point1 = QPointF(point1[0], point1[1])
            point2 = QPointF(point2[0], point2[1])
            qline = self.scene.addLine(QLineF(point1, point2))
            arrow_points, text_pos = _port_marker(port, is_subport)
            arrow_qpoly = QPolygonF([QPointF(p[0], p[1]) for p in arrow_points])
            port_scene_poly = self.scene.addPolygon(arrow_qpoly)
            # port_scene_poly.setRotation(port.orientation)
            # port_scene_poly.moveBy(port.center[0], port.center[1])
            port_shapes = [qline, port_scene_poly]
        qtext = self.scene.addText(str(port.name), self.portfont)
        qtext.setPos(QPointF(text_pos[0], text_pos[1]))
        qtext.setFlag(QGraphicsItem.ItemIgnoresTransformations)
        port_items = port_shapes + [qtext]

        if not is_subport:
            [shape.setPen(self.portpen) for shape in port_shapes]
            qtext.setDefaultTextColor(self.portfontcolor)
            self.portitems += port_items
        else:
            [shape.setPen(self.subportpen) for shape in port_shapes]
            qtext.setDefaultTextColor(self.subportfontcolor)
            self.subportitems += port_items

    #        self.portlabels.append(qtext)

    def add_aliases(self, aliases):
        for name, ref in aliases.items():
            qtext = self.scene.addText(str(name), self.portfont)
            x, y = ref.center
            qtext.setPos(QPointF(x, y))
            qtext.setFlag(QGraphicsItem.ItemIgnoresTransformations)
            self.aliasitems += [qtext]

    def set_port_visibility(self, visible=True):
        for item in self.portitems:
            item.setVisible(visible)
        self.ports_visible = visible

    def set_subport_visibility(self, visible=True):
        for item in self.subportitems:
            item.setVisible(visible)
        self.subports_visible = visible

    def set_alias_visibility(self, visible=True):
        for item in self.aliasitems:
            item.setVisible(visible)
        self.aliases_visible = visible

    def initialize(self):
        self.scene.clear()
        self.polygons = {}
        self.portitems = []
        self.subportitems = []
        self.aliasitems = []
        self.aliases_visible = True
        self.ports_visible = True
        self.subports_visible = True
        self.mouse_position = [0, 0]
        self.grid_size_snapped = 0
        self.setMouseTracking(True)
        self.scene_bounding_rect = None
        self.scene_polys = []
        self.scene_xmin = 0
        self.scene_xmax = 1
        self.scene_ymin = 0
        self.scene_ymax = 1

    def finalize(self):
        self.scene_bounding_rect = QRectF(
            QPointF(self.scene_xmin, self.scene_ymin),
            QPointF(self.scene_xmax, self.scene_ymax),
        )
        # self.scene_center = [self.scene_bounding_rect.center().x(), self.scene_bounding_rect.center().y()]
        self.scene_size = [
            self.scene_bounding_rect.width(),
            self.scene_bounding_rect.height(),
        ]
        self.create_grid()
        self.update_grid()

    # ==============================================================================
    #   Grid creation
    # ==============================================================================
    def update_grid(self):
        grid_pixels = 50
        grid_snaps = [1, 2, 4]

        # Number of pixels in the viewer
        view_width, _ = self.rect().width(), self.rect().height()
        # Rectangle of viewport in terms of scene coordinates
        r = self.mapToScene(self.rect()).boundingRect()
        width, height = r.width(), r.height()
        xmin, ymin, _, _ = r.x(), r.y(), r.x() + width, r.y() + height

        grid_size = grid_pixels * (width / view_width)
        exponent = np.floor(np.log10(grid_size))
        digits = round(grid_size / 10 ** (exponent), 2)
        digits_snapped = min(grid_snaps, key=lambda x: abs(x - digits))
        grid_size_snapped = digits_snapped * 10 ** (exponent)

        # Starting coordinates for gridlines
        x = round((xmin - 2 * width) / grid_size_snapped) * grid_size_snapped
        y = round((ymin - 2 * height) / grid_size_snapped) * grid_size_snapped

        for gl in self.gridlinesx:
            gl.setLine(x, -1e10, x, 1e10)
            x += grid_size_snapped
        for gl in self.gridlinesy:
            gl.setLine(-1e10, y, 1e10, y)
            y += grid_size_snapped
        self.grid_size_snapped = grid_size_snapped
        self.update_gridsize_label()

    def update_gridsize_label(self):
        self.gridsize_label.setText(f"grid size = {str(self.grid_size_snapped)}")
        self.gridsize_label.move(QPoint(5, self.height() - 25))

    def update_mouse_position_label(self):
        self.position_label.setText(
            "X = {:0.4f} / Y = {:0.4f}".format(
                self.mouse_position[0], self.mouse_position[1]
            )
        )
        self.position_label.move(QPoint(self.width() - 250, self.height() - 25))

    def update_help_label(self):
        self.help_label.setText('Press "?" key for help')
        self.help_label.move(QPoint(self.width() - 175, 0))

    def create_grid(self):
        self.gridlinesx = [
            self.scene.addLine(-10, -10, 10, 10, self.gridpen) for _ in range(300)
        ]

        self.gridlinesy = [
            self.scene.addLine(-10, -10, 10, 10, self.gridpen) for _ in range(300)
        ]

        self.update_grid()

    # ==============================================================================
    #  Mousewheel zoom, taken from http://stackoverflow.com/a/29026916
    # ==============================================================================
    def wheelEvent(self, event):
        # Zoom Factor
        zoom_percentage = 1.4

        # Set Anchors
        self.setTransformationAnchor(QGraphicsView.NoAnchor)
        self.setResizeAnchor(QGraphicsView.NoAnchor)

        # Save the scene pos
        oldPos = self.mapToScene(event.pos())

        # Zoom
        mousewheel_rotation = (
            event.angleDelta().y()
        )  # Typically = 120 on most mousewheels
        zoom_factor = zoom_percentage ** (mousewheel_rotation / 120)
        zoom_factor = np.clip(zoom_factor, 0.5, 2.0)

        # Check to make sure we're not overzoomed
        min_width = 0.01
        min_height = 0.01

        window_width = self.rect().width()
        window_height = self.rect().height()
        scene_upper_left_corner = self.mapToScene(QPoint(0, 0))
        scene_bottom_right_corner = self.mapToScene(QPoint(window_width, window_height))
        scene_width = (scene_bottom_right_corner - scene_upper_left_corner).x()
        scene_height = (scene_upper_left_corner - scene_bottom_right_corner).y()

        max_width = self.scene_bounding_rect.width() * 3
        max_height = self.scene_bounding_rect.height() * 3

        if ((scene_width > max_width) and (scene_height > max_height)) and (
            zoom_factor < 1
        ):
            pass
        elif scene_width >= min_width or scene_height >= min_height or zoom_factor <= 1:
            self.zoom_view(zoom_factor)

        # Get the new position and move scene to old position
        newPos = self.mapToScene(event.pos())
        delta = newPos - oldPos
        self.translate(delta.x(), delta.y())

        self.update_grid()

    def zoom_view(self, zoom_factor):
        old_center = self.mapToScene(self.rect().center())
        self.scale(zoom_factor, zoom_factor)
        self.centerOn(old_center)
        self.zoom_factor_total *= zoom_factor

    def resizeEvent(self, event):
        super(QGraphicsView, self).resizeEvent(event)
        if self.scene_bounding_rect is not None:
            self.reset_view()
        self.update_gridsize_label()
        self.update_mouse_position_label()
        self.update_help_label()

    def mousePressEvent(self, event):
        super(QGraphicsView, self).mousePressEvent(event)
        # ==============================================================================
        #  Zoom to rectangle, from
        #  https://wiki.python.org/moin/PyQt/Selecting%20a%20region%20of%20a%20widget
        # ==============================================================================
        if event.button() == Qt.RightButton:
            self._mousePressed = Qt.RightButton
            self._rb_origin = QPoint(event.pos())
            self.rubberBand.setGeometry(QRect(self._rb_origin, QSize()))
            self.rubberBand.show()
        # ==============================================================================
        # Mouse panning, taken from
        # http://stackoverflow.com/a/15043279
        # ==============================================================================
        elif event.button() == Qt.MidButton:
            self._mousePressed = Qt.MidButton
            self._mousePressedPos = event.pos()
            self.setCursor(QtCore.Qt.ClosedHandCursor)
            self._dragPos = event.pos()

    def mouseMoveEvent(self, event):
        super(QGraphicsView, self).mouseMoveEvent(event)

        # # Useful debug
        # try:
        #     self.debug_label.setText(str(itemsBoundingRect_nogrid().width()))
        # except Exception:
        #     print('Debug statement failed')

        # Update the X,Y label indicating where the mouse is on the geometry
        mouse_position = self.mapToScene(event.pos())
        self.mouse_position = [mouse_position.x(), mouse_position.y()]
        self.update_mouse_position_label()

        if not self._rb_origin.isNull() and self._mousePressed == Qt.RightButton:
            self.rubberBand.setGeometry(
                QRect(self._rb_origin, event.pos()).normalized()
            )

        # Middle-click-to-pan
        if self._mousePressed == Qt.MidButton:
            newPos = event.pos()
            diff = newPos - self._dragPos
            self._dragPos = newPos
            self.horizontalScrollBar().setValue(
                self.horizontalScrollBar().value() - diff.x()
            )
            self.verticalScrollBar().setValue(
                self.verticalScrollBar().value() - diff.y()
            )

    #            event.accept()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.RightButton:
            self.rubberBand.hide()
            rb_rect = QRect(self._rb_origin, event.pos())
            rb_center = rb_rect.center()
            rb_size = rb_rect.size()

            if abs(rb_size.width()) > 3 and abs(rb_size.height()) > 3:
                viewport_size = self.viewport().geometry().size()

                zoom_factor_x = abs(viewport_size.width() / rb_size.width())
                zoom_factor_y = abs(viewport_size.height() / rb_size.height())

                new_center = self.mapToScene(rb_center)

                zoom_factor = min(zoom_factor_x, zoom_factor_y)
                self.zoom_view(zoom_factor)
                self.centerOn(new_center)

            self.update_grid()

        if event.button() == Qt.MidButton:
            self.setCursor(Qt.ArrowCursor)
            self._mousePressed = None
            self.update_grid()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.reset_view()

        if event.key() == Qt.Key_F1:
            self.set_alias_visibility(not self.aliases_visible)

        if event.key() == Qt.Key_F2:
            self.set_port_visibility(not self.ports_visible)

        if event.key() == Qt.Key_F3:
            self.set_subport_visibility(not self.subports_visible)

        if event.key() == Qt.Key_Question:
            help_str = """
            Mouse control:
              Mousewheel: Zoom in and out
              Right-click & drag: Zoom to rectangle
              Middle-click & drag: Pan

            Keyboard shortcuts:
              Esc: Reset view
              F1: Show/hide alias names
              F2: Show/hide ports
              F3: Show/hide subports (ports in underlying references)
            """
            QMessageBox.about(self, "PHIDL Help", help_str)


def quickplot2(item_list, *args, **kwargs):
    """QT plot."""
    from gdsfactory.pdk import get_layer_views
    from gdsfactory.technology import LayerView

    if not qt_imported:
        raise ImportError(
            "quickplot2 tried to import PyQt5 but it failed. gdsfactory will"
            "still work but quickplot2() may not.  Try using"
            "quickplot() instead (based on matplotlib)"
        )

    global app
    if QCoreApplication.instance() is None:
        app = QApplication(sys.argv)
    if "viewer_window" not in globals():
        global viewer_window
        viewer_window = ViewerWindow()
    viewer = viewer_window.viewer
    viewer.initialize()
    if not isinstance(item_list, (list, tuple)):
        item_list = [item_list]
    LAYER_VIEWS = get_layer_views()
    for element in item_list:
        if isinstance(
            element,
            (
                Component,
                ComponentReference,
            ),
        ):
            # Draw polygons in the element
            polygons_spec = element.get_polygons(by_spec=True, depth=None)
            for key in sorted(polygons_spec):
                polygons = polygons_spec[key]
                layer_view = (
                    LAYER_VIEWS.get_from_tuple(key)
                    if key in LAYER_VIEWS
                    else LayerView(layer=key)
                )
                colors = layer_view.get_color_dict()
                viewer.add_polygons(
                    polygons,
                    fill_color=colors["fill_color"],
                    frame_color=colors["frame_color"],
                    alpha=layer_view.get_alpha(),
                )
            # If element is a Component, draw ports and aliases
            if isinstance(element, Component):
                for ref in element.references:
                    for port in ref.ports.values():
                        viewer.add_port(port, is_subport=True)
                for port in element.ports.values():
                    viewer.add_port(port)
                    viewer.add_aliases(element.named_references)
            # If element is a ComponentReference, draw ports as subports
            if isinstance(element, ComponentReference):
                for port in element.ports.values():
                    viewer.add_port(port, is_subport=True)
        elif isinstance(element, (Polygon)):
            layer_tuple = (element.layer, element.datatype)
            layer_view = (
                LAYER_VIEWS.get_from_tuple(layer_tuple)
                if layer_tuple in LAYER_VIEWS
                else LayerView(layer=layer_tuple)
            )
            colors = layer_view.get_color_dict()
            viewer.add_polygons(
                element.polygons,
                fill_color=colors["fill_color"],
                frame_color=colors["frame_color"],
                alpha=layer_view.get_alpha(),
            )
    viewer.finalize()
    viewer.reset_view()
    viewer_window.setVisible(True)
    viewer_window.show()
    viewer_window.raise_()
    return viewer


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    import gdsfactory as gf

    # set_quickplot_options(label_aliases=True, show_ports=False, show_subports=False)
    c = gf.components.straight_pn()
    c.plotqt()
    # c.plot()
    c.show()
    quickplot(c, show_ports=False, show_subports=False, label_aliases=True)
    plt.show()



import numpy as np


def free_spectral_range(n: float, length: float, wavelength: float) -> float:
    """Returns Free Spectral range in um.

    Args:
        n: index.
        length: in um.
        wavelength: in um.

    """
    return wavelength**2 / (n * length)


def directional_coupler_lc(
    wavelength_nm: int, n_eff_1: float, n_eff_2: float, power_ratio: float = 1.0
) -> float:
    """Calculates directional coupler coherence length (100% power transfer).

    Args:
        wavelength_nm: The wavelength in [nm] the directional coupler should operate at.
        n_eff_1: n_eff of the fundamental (even) supermode of the directional coupler.
        n_eff_2: n_eff of the first-order (odd) supermode of the directional coupler.
        power_ratio: p2/p1.

    """
    wavelength_m = wavelength_nm * 1.0e-9
    dn_eff = (n_eff_1 - n_eff_2).real
    lc_m = wavelength_m / (np.pi * dn_eff) * np.arcsin(np.sqrt(power_ratio))
    return lc_m * 1.0e6


def grating_coupler_period(
    wavelength: float,
    n_eff: float,
    n_clad: float,
    incidence_angle_deg: float,
    diffration_order: int = 1,
) -> float:
    """Returns period needed for a grating coupler.

    Args:
        wavelength: target wavelength for the grating coupler.
        n_eff: effective index of the mode of a waveguide with the width of the grating coupler.
        n_clad refractive index of the cladding.
        incidence_angle_deg: incidence angle the grating coupler should operate in degrees.
        diffration_order: grating order the coupler should work at. Default is 1st order (1).

    Returns:
        float: The period needed for the grating coupler in the same units as wavelength
    """
    k0 = 2.0 * np.pi / wavelength
    beta = n_eff.real * k0
    n_inc = n_clad

    return (2.0 * np.pi * diffration_order) / (
        beta - k0 * n_inc * np.sin(np.radians(incidence_angle_deg))
    )


if __name__ == "__main__":
    fsr = free_spectral_range(n=2.4, length=20, wavelength=1.55)
    print(fsr * 1e3)



"""From phidl tutorial.

# Flattening a Component

Sometimes you want to remove cell structure from a Component while keeping all
of the shapes/polygons intact and in place.

The Component.flatten() method returns a new flatten Component with all the
polygons inside the Component, and removes all the underlying references.
Also, if you specify the `single_layer` argument it will move all of the
polygons to that single layer.

"""
from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component


def test_flatten_device() -> Component:
    c = gf.Component("test_remap_layers")

    c.add_ref(gf.components.rectangle(size=(10, 1), layer=gf.LAYER.WG))
    c.add_ref(gf.components.rectangle(size=(10, 2), layer=gf.LAYER.SLAB90))
    c.add_ref(gf.components.rectangle(size=(10, 3), layer=gf.LAYER.SLAB150))

    assert len(c.references) == 3
    c2 = c.flatten()
    # c2 = c.flatten(single_layer=(34, 0))
    assert len(c2.references) == 0
    return c2


if __name__ == "__main__":
    c = test_flatten_device()
    c.show(show_ports=True)



from __future__ import annotations

from gdsfactory.read.from_dphox import from_dphox
from gdsfactory.read.from_gdspaths import from_gdsdir, from_gdspaths
from gdsfactory.read.from_np import from_np
from gdsfactory.read.from_phidl import from_gdstk, from_phidl
from gdsfactory.read.from_yaml import from_yaml, cell_from_yaml
from gdsfactory.read.from_yaml_template import cell_from_yaml_template
from gdsfactory.read.import_gds import import_gds, import_gds_raw

__all__ = [
    "from_dphox",
    "from_gdsdir",
    "from_gdspaths",
    "from_np",
    "from_phidl",
    "from_yaml",
    "from_gdstk",
    "cell_from_yaml",
    "cell_from_yaml_template",
    "import_gds",
    "import_gds_raw",
]



from __future__ import annotations

import pathlib
import tempfile
import uuid
from typing import Tuple

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentOrPath

valid_operations = ("xor", "not", "and", "or")


@gf.cell
def boolean_klayout(
    gdspath1: ComponentOrPath,
    gdspath2: ComponentOrPath,
    layer1: Tuple[int, int] = (1, 0),
    layer2: Tuple[int, int] = (1, 0),
    layer3: Tuple[int, int] = (2, 0),
    operation: str = "xor",
) -> Component:
    """Returns a boolean operation between two components Uses KLayout python API.

    Args:
        gdspath1: path to GDS or Component.
        gdspath2: path to GDS or Component.
        layer1: tuple for gdspath1.
        layer2: tuple for gdspath2.
        layer3: for the result of the operation.

    """
    import klayout.db as pya

    if operation not in valid_operations:
        raise ValueError(f"{operation} not in {valid_operations}")

    if isinstance(gdspath1, Component):
        gdspath1.flatten()
        gdspath1 = gdspath1.write_gds()

    if isinstance(gdspath2, Component):
        gdspath2.flatten()
        gdspath2 = gdspath2.write_gds()

    layout1 = pya.Layout()
    layout1.read(str(gdspath1))
    cell1 = layout1.top_cell()

    layout2 = pya.Layout()
    layout2.read(str(gdspath2))
    cell2 = layout2.top_cell()

    cellname = f"boolean_{str(uuid.uuid4())[:8]}"
    layout3 = pya.Layout()
    layout3_top = layout3.create_cell(cellname)

    a = pya.Region(cell1.begin_shapes_rec(layout1.layer(layer1[0], layer1[1])))
    b = pya.Region(cell2.begin_shapes_rec(layout2.layer(layer2[0], layer2[1])))

    if operation == "xor":
        result = a ^ b
    elif operation == "not":
        result = a - b
    elif operation == "and":
        result = a & b
    elif operation == "or":
        result = a | b

    layout3_top.shapes(layout3.layer(layer3[0], layer3[1])).insert(result)

    dirpath_build = pathlib.Path(tempfile.TemporaryDirectory().name)
    dirpath_build.mkdir(exist_ok=True, parents=True)
    gdspath = str(dirpath_build / f"{cellname}.gds")
    layout3.write(gdspath)
    return gf.import_gds(gdspath)


def _demo() -> None:
    import klayout.db as pya

    import gdsfactory as gf

    gdspath1 = gf.Component("ellipse1")
    gdspath1.add_ref(gf.components.ellipse(radii=[10, 5], layer=(1, 0)))

    gdspath2 = gf.Component("ellipse2")
    gdspath2.add_ref(gf.components.ellipse(radii=[11, 4], layer=(1, 0))).movex(4)

    layer1 = layer2 = (1, 0)
    layer3 = (1, 0)

    if isinstance(gdspath1, Component):
        gdspath1.flatten()
        gdspath1 = gdspath1.write_gds()

    if isinstance(gdspath2, Component):
        gdspath2.flatten()
        gdspath2 = gdspath2.write_gds()
        gf.show(gdspath2)

    layout1 = pya.Layout()
    layout1.read(str(gdspath1))
    cell1 = layout1.top_cell()

    layout2 = pya.Layout()
    layout2.read(str(gdspath2))
    cell2 = layout2.top_cell()

    layout3 = pya.Layout()
    layout3_top = layout3.create_cell("top")

    a = pya.Region(cell1.begin_shapes_rec(layout1.layer(layer1[0], layer1[1])))
    b = pya.Region(cell2.begin_shapes_rec(layout2.layer(layer2[0], layer2[1])))
    rxor = a ^ b
    layout3_top.shapes(layout3.layer(layer3[0], layer3[1])).insert(rxor)

    layout3_top.write("boolean.gds")
    gf.show("boolean.gds")


def _show_shapes() -> None:
    c1 = gf.components.ellipse(radii=[8, 8], layer=(1, 0))
    c2 = gf.components.ellipse(radii=[11, 4], layer=(1, 0))
    c3 = gf.Component()
    c3 << c1
    c3 << c2
    c3.show()


if __name__ == "__main__":
    # _show_shapes()
    c1 = gf.components.ellipse(radii=[8, 8], layer=(1, 0))
    c2 = gf.components.ellipse(radii=[11, 4], layer=(1, 0))
    c = boolean_klayout(c1, c2, operation="not")
    c.show(show_ports=True)



"""Add ports from pin markers or labels."""
from __future__ import annotations

from functools import partial
from typing import Optional, Tuple

import numpy as np
from numpy import arctan2, degrees, isclose

from gdsfactory.component import Component
from gdsfactory.port import Port, read_port_markers, sort_ports_clockwise
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import LayerSpec


def add_ports_from_markers_square(
    component: Component,
    pin_layer: LayerSpec = "DEVREC",
    port_layer: Optional[LayerSpec] = None,
    orientation: Optional[int] = 90,
    min_pin_area_um2: float = 0,
    max_pin_area_um2: float = 150 * 150,
    pin_extra_width: float = 0.0,
    port_names: Optional[Tuple[str, ...]] = None,
    port_name_prefix: Optional[str] = None,
    port_type: str = "optical",
) -> Component:
    """Add ports from square markers at the port center in port_layer.

    Args:
        component: to read polygons from and to write ports to.
        pin_layer: for port markers.
        port_layer: for the new created port.
        orientation: in degrees 90 north, 0 east, 180 west, 270 south.
        min_pin_area_um2: ignores pins with area smaller than min_pin_area_um2.
        max_pin_area_um2: ignore pins for area above certain size.
        pin_extra_width: 2*offset from pin to straight.
        port_names: names of the ports (defaults to {i}).
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical.
        port_type: optical, electrical.
    """
    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default
    port_markers = read_port_markers(component, (pin_layer,))
    port_names = port_names or [
        f"{port_name_prefix}{i+1}" for i in range(len(port_markers.polygons))
    ]
    layer = port_layer or pin_layer

    for port_name, p in zip(port_names, port_markers.polygons):
        (xmin, ymin), (xmax, ymax) = p.bounding_box()
        x, y = np.sum(p.bounding_box(), 0) / 2

        dy = snap_to_grid(ymax - ymin)
        dx = snap_to_grid(xmax - xmin)
        if dx == dy and max_pin_area_um2 > dx * dy > min_pin_area_um2:
            x = x
            y = y
            component.add_port(
                port_name,
                center=(x, y),
                width=dx - pin_extra_width,
                orientation=orientation,
                layer=layer,
            )
    return component


def add_ports_from_markers_center(
    component: Component,
    pin_layer: LayerSpec = "PORT",
    port_layer: Optional[LayerSpec] = None,
    inside: bool = False,
    tol: float = 0.1,
    pin_extra_width: float = 0.0,
    min_pin_area_um2: Optional[float] = None,
    max_pin_area_um2: float = 150.0 * 150.0,
    skip_square_ports: bool = False,
    xcenter: Optional[float] = None,
    ycenter: Optional[float] = None,
    port_name_prefix: Optional[str] = None,
    port_type: str = "optical",
    short_ports: bool = False,
    auto_rename_ports: bool = True,
    debug: bool = False,
) -> Component:
    """Add ports from rectangular pin markers.

    markers at port center, so half of the marker goes inside and half outside the port.

    guess port orientation from the component center (xcenter)

    Args:
        component: to read polygons from and to write ports to.
        pin_layer: GDS layer for maker [int, int].
        port_layer: for the new created port.
        inside: True-> markers  inside. False-> markers at center.
        tol: tolerance for comparing how rectangular is the pin.
        pin_extra_width: 2*offset from pin to straight.
        min_pin_area_um2: ignores pins with area smaller than min_pin_area_um2.
        max_pin_area_um2: ignore pins for area above certain size.
        skip_square_ports: skips square ports (hard to guess orientation).
        xcenter: for guessing orientation of rectangular ports.
        ycenter: for guessing orientation of rectangular ports.
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical ports.
        port_type: type of port (optical, electrical ...).
        short_ports: if the port is on the short side rather than the long side
        auto_rename_ports:
        debug: if True prints ports that are skipped.

    For inside=False the port location is at the middle of the PIN

    .. code::
           _______________
          |               |
          |               |
         |||             |||____  | pin_extra_width/2 > 0
         |||             |||
         |||             |||____
         |||             |||
          |      __       |
          |_____|__|______|
                |__|

    For inside=True all the pin is inside the port

    .. code::
           _______________
          |               |
          |               |
          |_              |
          | |             |
          |_|             |
          |               |
          |      __       |
          |_____|__|______|

    dx < dy: port is east or west
        x > xc: east
        x < xc: west

    dx > dy: port is north or south
        y > yc: north
        y < yc: south

    dx = dy
        x > xc: east
        x < xc: west
    """
    xc = xcenter or component.x
    yc = ycenter or component.y
    xmax = component.xmax
    xmin = component.xmin
    ymax = component.ymax
    ymin = component.ymin

    port_markers = read_port_markers(component, layers=(pin_layer,))
    layer = port_layer or pin_layer
    port_locations = []

    ports = {}

    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default

    for i, p in enumerate(port_markers.polygons):
        port_name = f"{port_name_prefix}{i+1}" if port_name_prefix else str(i)
        (xmin, ymin), (xmax, ymax) = p.bounding_box()
        x, y = np.sum(p.bounding_box(), 0) / 2

        dy = ymax - ymin
        dx = xmax - xmin

        if min_pin_area_um2 and dx * dy < min_pin_area_um2:
            if debug:
                print(f"skipping port at ({x}, {y}) with min_pin_area_um2 {dx * dy}")
            continue

        if max_pin_area_um2 and dx * dy > max_pin_area_um2:
            continue

        if skip_square_ports and snap_to_grid(dx) == snap_to_grid(dy):
            if debug:
                print(f"skipping square port at ({x}, {y})")
            continue

        pxmax = xmax
        pxmin = xmin
        pymax = ymax
        pymin = ymin

        orientation = 0

        if dy < dx if short_ports else dx < dy:
            if x > xc:  # east
                orientation = 0
                width = dy
                x = xmax if inside else x
            elif x < xc:  # west
                orientation = 180
                width = dy
                x = xmin if inside else x
        elif dy > dx if short_ports else dx > dy:
            if y > yc:  # north
                orientation = 90
                width = dx
                y = ymax if inside else y
            elif y < yc:  # south
                orientation = 270
                width = dx
                y = ymin if inside else y

        elif pxmax > xmax - tol:  # east
            orientation = 0
            width = dy
            x = xmax if inside else x
        elif pxmin < xmin + tol:  # west
            orientation = 180
            width = dy
            x = xmin if inside else x
        elif pymax > ymax - tol:  # north
            orientation = 90
            width = dx
            y = ymax if inside else y
        elif pymin < ymin + tol:  # south
            orientation = 270
            width = dx
            y = ymin if inside else y

        elif pxmax > xc:
            orientation = 0
            width = dy
            x = xmax if inside else x

        elif pxmax < xc:
            orientation = 180
            width = dy
            x = xmin if inside else x

        x = snap_to_grid(x)
        y = snap_to_grid(y)
        width = np.round(width - pin_extra_width, 3)

        if (x, y) not in port_locations:
            port_locations.append((x, y))
            ports[port_name] = Port(
                name=port_name,
                center=(x, y),
                width=width,
                orientation=orientation,
                layer=layer,
                port_type=port_type,
            )

    ports = sort_ports_clockwise(ports)

    for port_name, port in ports.items():
        if port_name in component.ports:
            component_ports = list(component.ports.keys())
            raise ValueError(
                f"port {port_name!r} already in {component_ports}. "
                "You can pass a port_name_prefix to add it with a different name."
            )

        else:
            component.add_port(name=port_name, port=port)
    if auto_rename_ports:
        component.auto_rename_ports()
    return component


add_ports_from_markers_inside = partial(add_ports_from_markers_center, inside=True)


def add_ports_from_labels(
    component: Component,
    port_width: float,
    port_layer: LayerSpec,
    xcenter: Optional[float] = None,
    port_name_prefix: Optional[str] = None,
    port_type: str = "optical",
    get_name_from_label: bool = False,
    layer_label: Optional[LayerSpec] = None,
    fail_on_duplicates: bool = False,
    port_orientation: Optional[float] = None,
    guess_port_orientation: bool = True,
) -> Component:
    """Add ports from labels.

    Assumes that all ports have a label at the port center.
    because labels do not have width, you have to manually specify the ports width

    Args:
        component: to read polygons from and to write ports to.
        port_width: for ports.
        port_layer: for the new created port.
        xcenter: center of the component, for guessing port orientation.
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical.
        port_type: optical, electrical.
        layer_label:
        fail_on_duplicates: raises ValueError for duplicated port names.
            if False adds incremental suffix (1, 2 ...) to port name.
        port_orientation: None for electrical ports.
        guess_port_orientation: assumes right: 0, left: 180, top: 90, bot: 270.
    """
    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default
    yc = component.y

    port_name_to_index = {}

    xc = xcenter or component.x
    for i, label in enumerate(component.labels):
        x, y = label.origin

        if layer_label and (
            layer_label[0] != label.layer or layer_label[1] != label.texttype
        ):
            continue

        if get_name_from_label:
            port_name = label.text
        else:
            port_name = f"{port_name_prefix}{i+1}" if port_name_prefix else i

        orientation = port_orientation

        if guess_port_orientation:
            if x > xc:  # east
                orientation = 0
            elif x < xc:  # west
                orientation = 180
            elif y > yc:  # north
                orientation = 90
            elif y < yc:  # south
                orientation = 270

        if fail_on_duplicates and port_name in component.ports:
            component_ports = list(component.ports.keys())
            raise ValueError(
                f"port {port_name!r} already in {component_ports}. "
                "You can pass a port_name_prefix to add it with a different name."
            )
        if get_name_from_label and port_name in component.ports:
            port_name_to_index[label.text] = (
                port_name_to_index[label.text] + 1
                if label.text in port_name_to_index
                else 1
            )
            port_name = f"{label.text}{port_name_to_index[label.text]}"

        component.add_port(
            name=port_name,
            center=(x, y),
            width=port_width,
            orientation=orientation,
            port_type=port_type,
            layer=port_layer,
        )
    return component


def add_ports_from_siepic_pins(
    component: Component,
    pin_layer_optical: LayerSpec = "PORT",
    port_layer_optical: Optional[LayerSpec] = None,
    pin_layer_electrical: LayerSpec = "PORTE",
    port_layer_electrical: Optional[LayerSpec] = None,
) -> Component:
    """Add ports from SiEPIC-type cells, where the pins are defined as paths.

    Looks for label, path pairs.

    Args:
        component: component.
        pin_layer_optical: layer for optical pins.
        port_layer_optical: layer for optical ports.
        pin_layer_electrical: layer for electrical pins.
        port_layer_electrical: layer for electrical ports.
    """
    pin_layers = {"optical": pin_layer_optical, "electrical": pin_layer_electrical}

    import gdsfactory as gf

    pin_layer_optical = gf.get_layer(pin_layer_optical)
    port_layer_optical = gf.get_layer(port_layer_optical)
    pin_layer_electrical = gf.get_layer(pin_layer_electrical)
    port_layer_electrical = gf.get_layer(port_layer_electrical)

    c = component
    labels = c.get_labels()

    for path in c.paths:
        p1, p2 = path.spine()

        path_layers = list(zip(path.layers, path.datatypes))

        # Find the center of the path
        center = (p1 + p2) / 2

        # Find the label closest to the pin
        label = None
        for i, _label in enumerate(labels):
            if (
                all(isclose(_label.origin, center))
                or all(isclose(_label.origin, p1))
                or all(isclose(_label.origin, p2))
            ):
                label = _label
                labels.pop(i)
        if label is None:
            print(
                f"Warning: label not found for path: in center={center} p1={p1} p2={p2}"
            )
            continue
        if pin_layer_optical in path_layers:
            port_type = "optical"
            port_layer = port_layer_optical or None
        elif pin_layer_electrical in path_layers:
            port_type = "electrical"
            port_layer = port_layer_electrical or None
        else:
            continue

        port_name = str(label.text)

        # If the port name is already used, add a number to it
        i = 1
        while port_name in c.ports:
            port_name += f"_{i}"

        angle = round(degrees(arctan2(p2[1] - p1[1], p2[0] - p1[0])) % 360)

        port = Port(
            name=port_name,
            center=center,
            width=path.widths()[0][0],
            orientation=angle,
            layer=port_layer or pin_layers[port_type],
            port_type=port_type,
        )
        c.add_port(port)
    return c
